import warnings

import numpy as np
from numpy.testing import (
    assert_,
    assert_array_almost_equal,
    assert_equal,
    assert_raises,
    assert_warns,
)
import scipy.special as sps

from dipy.core.gradients import generate_bvecs, gradient_table
from dipy.denoise.localpca import (
    _pca_classifier,
    compute_num_samples,
    compute_patch_size,
    compute_suggested_patch_radius,
    create_patch_radius_arr,
    dimensionality_problem_message,
    genpca,
    localpca,
    mppca,
)
from dipy.sims.voxel import multi_tensor
from dipy.testing.decorators import set_random_number_generator


def setup_module():
    global gtab

    # generate a gradient table for phantom data
    directions8 = generate_bvecs(8)
    directions30 = generate_bvecs(30)
    directions60 = generate_bvecs(60)
    # Create full dataset parameters
    # (6 b-values = 0, 8 directions for b-value 300, 30 directions for b-value
    # 1000 and 60 directions for b-value 2000)
    bvals = np.hstack(
        (np.zeros(6), 300 * np.ones(8), 1000 * np.ones(30), 2000 * np.ones(60))
    )
    bvecs = np.vstack((np.zeros((6, 3)), directions8, directions30, directions60))
    gtab = gradient_table(bvals, bvecs=bvecs)


def rfiw_phantom(gtab, snr=None, rng=None):
    """rectangle fiber immersed in water"""
    # define voxel index
    slice_ind = np.zeros((10, 10, 8))
    slice_ind[4:7, 4:7, :] = 1
    slice_ind[4:7, 7, :] = 2
    slice_ind[7, 7, :] = 3
    slice_ind[7, 4:7, :] = 4
    slice_ind[7, 3, :] = 5
    slice_ind[4:7, 3, :] = 6
    slice_ind[3, 3, :] = 7
    slice_ind[3, 4:7, :] = 8
    slice_ind[3, 7, :] = 9

    # Define tissue diffusion parameters
    # Restricted diffusion
    ADr = 0.99e-3
    RDr = 0.0
    # Hindered diffusion
    ADh = 2.26e-3
    RDh = 0.87e-3
    # S0 value for tissue
    S1 = 50
    # Fraction between Restricted and Hindered diffusion
    fia = 0.51

    # Define water diffusion
    Dwater = 3e-3
    S2 = 100  # S0 value for water

    # Define tissue volume fraction for each voxel type (in index order)
    f = np.array([0.0, 1.0, 0.6, 0.18, 0.30, 0.15, 0.50, 0.35, 0.70, 0.42])

    # Define S0 for each voxel (in index order)
    S0 = S1 * f + S2 * (1 - f)

    # multi tensor simulations assume that each water pull as constant S0
    # since I am assuming that tissue and water voxels have different S0,
    # tissue volume fractions have to be adjusted to the measured f values when
    # constant S0 are assumed constant. Doing this correction, simulations will
    # be analogous to simulates that S0 are different for each media. (For more
    # details on this contact the phantom designer)
    f1 = f * S1 / S0

    mevals = np.array([[ADr, RDr, RDr], [ADh, RDh, RDh], [Dwater, Dwater, Dwater]])
    angles = [(0, 0, 1), (0, 0, 1), (0, 0, 1)]
    DWI = np.zeros(slice_ind.shape + (gtab.bvals.size,))
    for i in range(10):
        fractions = [f1[i] * fia * 100, f1[i] * (1 - fia) * 100, (1 - f1[i]) * 100]
        sig, direction = multi_tensor(
            gtab, mevals, S0=S0[i], angles=angles, fractions=fractions, snr=None
        )
        DWI[slice_ind == i, :] = sig

    if snr is None:
        return DWI
    else:
        if rng is None:
            rng = np.random.default_rng()
        sigma = S2 * 1.0 / snr
        n1 = rng.normal(0, sigma, size=DWI.shape)
        n2 = rng.normal(0, sigma, size=DWI.shape)
        return [
            np.sqrt((DWI / np.sqrt(2) + n1) ** 2 + (DWI / np.sqrt(2) + n2) ** 2),
            sigma,
        ]


def test_lpca_static():
    S0 = 100 * np.ones((20, 20, 20, 20), dtype="f8")
    S0ns = localpca(S0, sigma=np.ones((20, 20, 20), dtype=np.float64))
    assert_array_almost_equal(S0, S0ns)


@set_random_number_generator()
def test_lpca_random_noise(rng):
    S0 = 100 + 2 * rng.standard_normal((22, 23, 30, 20))
    S0ns = localpca(S0, sigma=np.std(S0))

    assert_(S0ns.min() > S0.min())
    assert_(S0ns.max() < S0.max())
    assert_equal(np.round(S0ns.mean()), 100)


@set_random_number_generator()
def test_lpca_boundary_behaviour(rng):
    # check is first slice is getting denoised or not ?
    S0 = 100 * np.ones((20, 20, 20, 20), dtype="f8")
    S0[:, :, 0, :] = S0[:, :, 0, :] + 2 * rng.standard_normal((20, 20, 20))
    S0_first = S0[:, :, 0, :]
    S0ns = localpca(S0, sigma=np.std(S0))
    S0ns_first = S0ns[:, :, 0, :]
    rmses = np.sum(np.abs(S0ns_first - S0_first)) / (100.0 * 20.0 * 20.0 * 20.0)

    # shows that S0n_first is not very close to S0_first
    assert_(rmses > 0.0001)
    assert_equal(np.round(S0ns_first.mean()), 100)

    rmses = np.sum(np.abs(S0ns_first - S0_first)) / (100.0 * 20.0 * 20.0 * 20.0)

    # shows that S0n_first is not very close to S0_first
    assert_(rmses > 0.0001)
    assert_equal(np.round(S0ns_first.mean()), 100)


@set_random_number_generator()
def test_lpca_rmse(rng):
    S0_w_noise = 100 + 2 * rng.standard_normal((22, 23, 30, 20))
    rmse_w_noise = np.sqrt(np.mean((S0_w_noise - 100) ** 2))
    S0_denoised = localpca(S0_w_noise, sigma=np.std(S0_w_noise))
    rmse_denoised = np.sqrt(np.mean((S0_denoised - 100) ** 2))
    # Denoising should always improve the RMSE:
    assert_(rmse_denoised < rmse_w_noise)


@set_random_number_generator()
def test_lpca_sharpness(rng):
    S0 = np.ones((30, 30, 30, 20), dtype=np.float64) * 100
    S0[10:20, 10:20, 10:20, :] = 50
    S0[20:30, 20:30, 20:30, :] = 0
    S0 = S0 + 20 * rng.standard_normal((30, 30, 30, 20))
    S0ns = localpca(S0, sigma=20.0)
    # check the edge gradient
    edgs = np.abs(np.mean(S0ns[8, 10:20, 10:20] - S0ns[12, 10:20, 10:20]) - 50)
    assert_(edgs < 2)


def test_lpca_dtype():
    # If out_dtype is not specified, we retain the original precision:
    S0 = 200 * np.ones((20, 20, 20, 3), dtype=np.float64)
    S0ns = localpca(S0, sigma=1)
    assert_equal(S0.dtype, S0ns.dtype)

    S0 = 200 * np.ones((20, 20, 20, 20), dtype=np.uint16)
    S0ns = localpca(S0, sigma=np.ones((20, 20, 20)))
    assert_equal(S0.dtype, S0ns.dtype)

    # If we set out_dtype, we get what we asked for:
    S0 = 200 * np.ones((20, 20, 20, 20), dtype=np.uint16)
    S0ns = localpca(S0, sigma=np.ones((20, 20, 20)), out_dtype=np.float32)
    assert_equal(np.float32, S0ns.dtype)

    # If we set a few entries to zero, this induces negative entries in the
    # Resulting denoised array:
    S0[5:8, 5:8, 5:8] = 0
    # But if we should always get all non-negative results:
    S0ns = localpca(S0, sigma=np.ones((20, 20, 20)), out_dtype=np.uint16)
    assert_(np.all(S0ns >= 0))
    # And no wrap-around to crazy high values:
    assert_(np.all(S0ns <= 200))


def test_lpca_wrong():
    S0 = np.ones((20, 20))
    assert_raises(ValueError, localpca, S0, sigma=1)


@set_random_number_generator()
def test_phantom(rng):
    DWI_clean = rfiw_phantom(gtab, snr=None, rng=rng)
    DWI, sigma = rfiw_phantom(gtab, snr=30, rng=rng)
    # To test without Rician correction
    temp = (DWI_clean / sigma) ** 2
    DWI_clean_wrc = (
        sigma
        * np.sqrt(np.pi / 2)
        * np.exp(-0.5 * temp)
        * (
            (1 + 0.5 * temp) * sps.iv(0, 0.25 * temp)
            + 0.5 * temp * sps.iv(1, 0.25 * temp)
        )
        ** 2
    )

    DWI_den = localpca(DWI, sigma=sigma, patch_radius=3)
    rmse_den = np.sum(np.abs(DWI_clean - DWI_den)) / np.sum(np.abs(DWI_clean))
    rmse_noisy = np.sum(np.abs(DWI_clean - DWI)) / np.sum(np.abs(DWI_clean))

    rmse_den_wrc = np.sum(np.abs(DWI_clean_wrc - DWI_den)) / np.sum(
        np.abs(DWI_clean_wrc)
    )
    rmse_noisy_wrc = np.sum(np.abs(DWI_clean_wrc - DWI)) / np.sum(np.abs(DWI_clean_wrc))

    assert_(np.max(DWI_clean) / sigma < np.max(DWI_den) / sigma)
    assert_(np.max(DWI_den) / sigma < np.max(DWI) / sigma)
    assert_(rmse_den < rmse_noisy)
    assert_(rmse_den_wrc < rmse_noisy_wrc)

    # Check if the results of different PCA methods (eig, svd) are similar
    DWI_den_svd = localpca(DWI, sigma=sigma, pca_method="svd", patch_radius=3)
    assert_array_almost_equal(DWI_den, DWI_den_svd)

    assert_raises(ValueError, localpca, DWI, sigma=sigma, pca_method="empty")

    # Try this with a sigma volume, instead of a scalar
    sigma_vol = sigma * np.ones(DWI.shape[:-1])
    mask = np.zeros_like(DWI, dtype=bool)[..., 0]
    mask[2:-2, 2:-2, 2:-2] = True
    DWI_den = localpca(DWI, sigma=sigma_vol, mask=mask, patch_radius=3)
    DWI_clean_masked = DWI_clean.copy()
    DWI_clean_masked[~mask] = 0
    DWI_masked = DWI.copy()
    DWI_masked[~mask] = 0
    rmse_den = np.sum(np.abs(DWI_clean_masked - DWI_den)) / np.sum(
        np.abs(DWI_clean_masked)
    )
    rmse_noisy = np.sum(np.abs(DWI_clean_masked - DWI_masked)) / np.sum(
        np.abs(DWI_clean_masked)
    )

    DWI_clean_wrc_masked = DWI_clean_wrc.copy()
    DWI_clean_wrc_masked[~mask] = 0
    rmse_den_wrc = np.sum(np.abs(DWI_clean_wrc_masked - DWI_den)) / np.sum(
        np.abs(DWI_clean_wrc_masked)
    )
    rmse_noisy_wrc = np.sum(np.abs(DWI_clean_wrc_masked - DWI_masked)) / np.sum(
        np.abs(DWI_clean_wrc_masked)
    )

    assert_(np.max(DWI_clean) / sigma < np.max(DWI_den) / sigma)
    assert_(np.max(DWI_den) / sigma < np.max(DWI) / sigma)
    assert_(rmse_den < rmse_noisy)
    assert_(rmse_den_wrc < rmse_noisy_wrc)


@set_random_number_generator()
def test_lpca_ill_conditioned(rng):
    DWI, sigma = rfiw_phantom(gtab, snr=30, rng=rng)
    for patch_radius in [1, [1, 1, 1]]:
        assert_warns(UserWarning, localpca, DWI, sigma=sigma, patch_radius=patch_radius)


@set_random_number_generator()
def test_lpca_radius_wrong_shape(rng):
    DWI, sigma = rfiw_phantom(gtab, snr=30, rng=rng)
    for patch_radius in [[2, 2], [2, 2, 2, 2]]:
        assert_raises(ValueError, localpca, DWI, sigma=sigma, patch_radius=patch_radius)


@set_random_number_generator()
def test_lpca_sigma_wrong_shape(rng):
    DWI, sigma = rfiw_phantom(gtab, snr=30, rng=rng)
    # If sigma is 3D but shape is not like DWI.shape[:-1], an error is raised:
    sigma = np.zeros((DWI.shape[0], DWI.shape[1] + 1, DWI.shape[2]))
    assert_raises(ValueError, localpca, DWI, sigma=sigma)


@set_random_number_generator()
def test_lpca_no_gtab_no_sigma(rng):
    DWI, sigma = rfiw_phantom(gtab, snr=30, rng=rng)
    assert_raises(ValueError, localpca, DWI, sigma=None, mask=None)


@set_random_number_generator()
def test_pca_classifier(rng):
    # Produce small phantom with well aligned single voxels and ground truth
    # snr = 50, i.e signal std = 0.02 (Gaussian noise)
    std_gt = 0.02
    S0 = 1.0
    ndir = gtab.bvals.size
    signal_test = np.zeros((5, 5, 5, ndir))
    mevals = np.array([[0.99e-3, 0.0, 0.0], [2.26e-3, 0.87e-3, 0.87e-3]])
    sig, direction = multi_tensor(
        gtab, mevals, S0=S0, angles=[(0, 0, 1), (0, 0, 1)], fractions=(50, 50), snr=None
    )
    signal_test[..., :] = sig
    noise = std_gt * rng.standard_normal((5, 5, 5, ndir))
    dwi_test = signal_test + noise

    # Compute eigenvalues
    X = dwi_test.reshape(125, ndir)
    M = np.mean(X, axis=0)
    X = X - M
    [L, W] = np.linalg.eigh(np.dot(X.T, X) / 125)

    # Find number of noise related eigenvalues
    var, c = _pca_classifier(L, 125)
    std = np.sqrt(var)

    # Expected number of signal components is 0 because the phantom only has
    # one voxel type and that information is captured by the mean of X.
    # Therefore, expected noise components should be equal to size of L.
    # To allow some margin of error let's assess if c is higher than
    # L.size - 3.
    assert_(c > L.size - 3)

    # Let's check if noise std estimate as an error less than 5%
    std_error = abs(std - std_gt) / std_gt * 100
    assert_(std_error < 5)


@set_random_number_generator()
def test_mppca_in_phantom(rng):
    DWIgt = rfiw_phantom(gtab, snr=None, rng=rng)
    std_gt = 0.02
    noise = std_gt * rng.standard_normal(DWIgt.shape)
    DWInoise = DWIgt + noise

    # patch radius (2: #samples > #features, 1: #samples < #features)
    for PR in [2, 1]:
        if PR == 1:
            patch_radius_arr = create_patch_radius_arr(DWInoise, PR)
            patch_size = compute_patch_size(patch_radius_arr)
            num_samples = compute_num_samples(patch_size)
            spr = compute_suggested_patch_radius(DWInoise, patch_size)
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore",
                    message=dimensionality_problem_message(DWInoise, num_samples, spr),
                    category=UserWarning,
                )
                DWIden = mppca(DWInoise, patch_radius=PR)
        else:
            DWIden = mppca(DWInoise, patch_radius=PR)

        # Test if denoised data is closer to ground truth than noisy data
        rmse_den = np.sum(np.abs(DWIgt - DWIden)) / np.sum(np.abs(DWIgt))
        rmse_noisy = np.sum(np.abs(DWIgt - DWInoise)) / np.sum(np.abs(DWIgt))
        assert_(rmse_den < rmse_noisy)


@set_random_number_generator()
def test_create_patch_radius_arr(rng):
    shape = (10, 10, 8, 104)
    arr = rng.standard_normal(shape)
    pr = 2
    expected_val = np.asarray([2, 2, 2])
    obtained_val = create_patch_radius_arr(arr, pr)
    assert np.array_equal(obtained_val, expected_val)


def test_compute_patch_size():
    patch_radius = 1
    expected_val = 3
    obtained_val = compute_patch_size(patch_radius)
    assert obtained_val == expected_val

    patch_radius = 2
    expected_val = 5
    obtained_val = compute_patch_size(patch_radius)
    assert obtained_val == expected_val


def test_compute_num_samples():
    patch_size = np.asarray([5, 5, 5])
    expected_val = 125
    obtained_val = compute_num_samples(patch_size)
    assert obtained_val == expected_val


@set_random_number_generator()
def test_compute_suggested_patch_radius(rng):
    shape = (10, 10, 8, 104)
    arr = rng.standard_normal(shape)
    patch_size = [3, 3, 3]
    expected_val = 2
    obtained_val = compute_suggested_patch_radius(arr, patch_size)
    assert obtained_val == expected_val

    patch_size = [5, 5, 5]
    obtained_val = compute_suggested_patch_radius(arr, patch_size)
    assert obtained_val == expected_val


@set_random_number_generator()
def test_mppca_returned_sigma(rng):
    DWIgt = rfiw_phantom(gtab, snr=None, rng=rng)
    std_gt = 0.02
    noise = std_gt * rng.standard_normal(DWIgt.shape)
    DWInoise = DWIgt + noise

    # patch radius (2: #samples > #features, 1: #samples < #features)
    for PR in [2, 1]:
        # Case that sigma is estimated using mpPCA
        if PR == 1:
            patch_radius_arr = create_patch_radius_arr(DWInoise, PR)
            patch_size = compute_patch_size(patch_radius_arr)
            num_samples = compute_num_samples(patch_size)
            spr = compute_suggested_patch_radius(DWInoise, patch_size)

        if PR == 1:
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore",
                    message=dimensionality_problem_message(DWInoise, num_samples, spr),
                    category=UserWarning,
                )
                DWIden0, sigma = mppca(DWInoise, patch_radius=PR, return_sigma=True)
        else:
            DWIden0, sigma = mppca(DWInoise, patch_radius=PR, return_sigma=True)
        msigma = np.mean(sigma)
        std_error = abs(msigma - std_gt) / std_gt * 100

        # if #noise_eigenvals >> #signal_eigenvals, variance should be well estimated
        # this is more likely achieved if #samples > #features
        if PR > 1:
            assert_(std_error < 5)
        else:  # otherwise, variance estimate may be wrong
            pass

        # Case that sigma is inputted (sigma outputted should be the same as the
        # one inputted)
        if PR == 1:
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore",
                    message=dimensionality_problem_message(DWInoise, num_samples, spr),
                    category=UserWarning,
                )
                DWIden1, rsigma = genpca(
                    DWInoise,
                    sigma=sigma,
                    tau_factor=None,
                    patch_radius=PR,
                    return_sigma=True,
                )
        else:
            DWIden1, rsigma = genpca(
                DWInoise,
                sigma=sigma,
                tau_factor=None,
                patch_radius=PR,
                return_sigma=True,
            )

        assert_array_almost_equal(rsigma, sigma)

        # DWIden1 should be very similar to DWIden0
        rmse_den = np.sum(np.abs(DWIden1 - DWIden0)) / np.sum(np.abs(DWIden0))
        rmse_ref = np.sum(np.abs(DWIden1 - DWIgt)) / np.sum(np.abs(DWIgt))
        assert_(rmse_den < rmse_ref)



#!/usr/bin/env python3
"""Script to auto-generate our API docs.
"""

# stdlib imports
from os.path import join as pjoin
import re
import sys

# local imports
from apigen import ApiDocWriter

# version comparison
from packaging.version import Version

# *****************************************************************************


def abort(error):
    print(f'*WARNING* API documentation not generated: {error}')
    exit()


if __name__ == '__main__':
    package = sys.argv[1]
    outdir = sys.argv[2]
    try:
        other_defines = sys.argv[3]
    except IndexError:
        other_defines = True
    else:
        other_defines = other_defines in ('True', 'true', '1')

    # Check that the package is available. If not, the API documentation is not
    # (re)generated and existing API documentation sources will be used.

    try:
        __import__(package)
    except ImportError as e:
        abort(f"Can not import {package}")

    # NOTE: with the new versioning scheme, this check is not needed anymore
    # Also, this might be needed if we do not use spin to generate the docs

    # module = sys.modules[package]

    # # Check that the source version is equal to the installed
    # # version. If the versions mismatch the API documentation sources
    # # are not (re)generated. This avoids automatic generation of documentation
    # # for older or newer versions if such versions are installed on the system.

    # installed_version = Version(module.__version__)
    # info_file = pjoin('..', package, 'info.py')
    # info_lines = open(info_file).readlines()
    # source_version = '.'.join([v.split('=')[1].strip(" '\n.")
    #                            for v in info_lines if re.match(
    #                                    '^_version_(major|minor|micro|extra)', v
    #                                    )]).strip('.')
    # source_version = Version(source_version)
    # print('***', source_version)

    # if source_version != installed_version:
    #     print('***', installed_version)
    #     abort("Installed version does not match source version")

    docwriter = ApiDocWriter(package, rst_extension='.rst',
                             other_defines=other_defines)
    docwriter.package_skip_patterns += [r'\.tracking\.interfaces.*$',
                                        r'\.tracking\.gui_tools.*$',
                                        r'.*test.*$',
                                        r'^\.utils.*',
                                        r'\.stats\.resampling.*$',
                                        r'\.info.*$',
                                        r'\.__config__.*$',
                                        ]
    docwriter.object_skip_patterns += [
        r'.*FetcherError.*$',
        r'.*urlopen.*',
        r'.*add_callback.*',
        r'.*Logger.*',
        r'.*logger.*',
        # Global variable. Must be enable in the future when using Typing.
        r'.*hemi_icosahedron.*',
        r'.*unit_octahedron.*',
        r'.*unit_icosahedron.*',
        r'.*default_sphere.*',
        r'.*small_sphere.*',
        r'.*icosahedron_faces.*',
        r'.*icosahedron_vertices.*',
        r'.*octahedron_faces.*',
        r'.*octahedron_vertices.*',
        r'.*diffusion_evals.*',
        r'.*DATA_DIR.*',
        r'.*RegistrationStages.*',
        r'.*VerbosityLevels.*',
    ]
    docwriter.write_api_docs(outdir)
    docwriter.write_index(outdir, 'index', relative_to=outdir)
    print(f'{len(docwriter.written_modules)} files written')



"""Test file for Parallel Transport Tracking Algorithm."""

import warnings

import numpy as np
import numpy.testing as npt

from dipy.core.sphere import unit_octahedron
from dipy.data import default_sphere, get_fnames
from dipy.direction import PTTDirectionGetter
from dipy.io.image import load_nifti
from dipy.reconst.shm import (
    SphHarmFit,
    SphHarmModel,
    descoteaux07_legacy_msg,
    sh_to_sf,
    tournier07_legacy_msg,
)
from dipy.tracking.local_tracking import LocalTracking
from dipy.tracking.stopping_criterion import BinaryStoppingCriterion
from dipy.tracking.streamline import Streamlines


def test_ptt_tracking():
    # Test PTT direction getter generate 100 streamlines with more than 1 pts.
    fod_fname, seed_coordinates_fname, _ = get_fnames(name="ptt_minimal_dataset")
    fod, affine = load_nifti(fod_fname)
    seed_coordinates = np.loadtxt(seed_coordinates_fname)[:10, :]
    sf = sh_to_sf(
        fod, default_sphere, basis_type="tournier07", sh_order_max=8, legacy=False
    )
    sf[sf < 0] = 0
    sc = BinaryStoppingCriterion(np.ones(fod.shape[:3]))

    dg_default = PTTDirectionGetter.from_pmf(sf, sphere=default_sphere, max_angle=20)
    dg_count2 = PTTDirectionGetter.from_pmf(
        sf, sphere=default_sphere, max_angle=20, probe_count=2, probe_radius=0.2
    )
    dg_quality10 = PTTDirectionGetter.from_pmf(
        sf, sphere=default_sphere, max_angle=20, probe_quality=10
    )
    dg_length2 = PTTDirectionGetter.from_pmf(
        sf, sphere=default_sphere, max_angle=20, probe_length=2
    )

    for dg in [dg_default, dg_count2, dg_quality10, dg_length2]:
        streamline_generator = LocalTracking(
            direction_getter=dg,
            step_size=0.2,
            stopping_criterion=sc,
            seeds=seed_coordinates,
            affine=affine,
        )
        streamlines = Streamlines(streamline_generator)
        npt.assert_equal(len(streamlines), 10)
        npt.assert_(np.all([len(s) > 1 for s in streamlines]))

    # Test with zeros pmf
    dg = PTTDirectionGetter.from_pmf(
        np.zeros(sf.shape), sphere=default_sphere, max_angle=20
    )
    streamline_generator = LocalTracking(
        direction_getter=dg,
        step_size=0.2,
        stopping_criterion=sc,
        seeds=seed_coordinates,
        affine=affine,
    )
    streamlines = Streamlines(streamline_generator)
    npt.assert_equal(len(streamlines), 10)
    npt.assert_(np.all([len(s) == 1 for s in streamlines]))

    # Test with maximum length reach
    dg = PTTDirectionGetter.from_pmf(sf, sphere=default_sphere, max_angle=20)
    streamline_generator = LocalTracking(
        direction_getter=dg,
        step_size=0.2,
        stopping_criterion=sc,
        seeds=seed_coordinates,
        affine=affine,
        maxlen=1,
        minlen=1,
    )
    streams = Streamlines(streamline_generator)
    npt.assert_almost_equal(
        np.linalg.norm(streams[0][0] - streams[0][1]), 0.2, decimal=1
    )
    npt.assert_equal(len(streams), 10)
    npt.assert_(np.all([len(s) <= 3 for s in streams]))

    streamline_generator = LocalTracking(
        direction_getter=dg,
        step_size=0.2,
        stopping_criterion=sc,
        seeds=seed_coordinates,
        affine=affine,
        maxlen=2,
        fixedstep=False,
    )

    # Check fixedstep ValueError
    npt.assert_raises(ValueError, Streamlines, streamline_generator)


def test_PTTDirectionGetter():
    # Test the constructors and errors of the PTTDirectionGetter
    class SillyModel(SphHarmModel):
        def fit(self, data, mask=None):
            coeff = np.zeros(data.shape[:-1] + (15,))
            return SphHarmFit(self, coeff, mask=None)

    silly_model = SillyModel(gtab=None)
    data = np.zeros((3, 3, 3, 7))
    fit = silly_model.fit(data)
    point = np.zeros(3)
    dir = unit_octahedron.vertices[0].copy()

    # Make ptt_dg from shm_coeffs
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        dg = PTTDirectionGetter.from_shcoeff(fit.shm_coeff, 90, unit_octahedron)
    npt.assert_equal(dg.get_direction(point, dir), 1)

    # Make ptt_dg from pmf
    pmf = np.zeros((3, 3, 3, unit_octahedron.theta.shape[0]))
    dg = PTTDirectionGetter.from_pmf(pmf, 90, unit_octahedron)
    npt.assert_equal(dg.get_direction(point, dir), 1)

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", message=tournier07_legacy_msg, category=PendingDeprecationWarning
        )

        # Check probe_length ValueError
        npt.assert_raises(
            ValueError,
            PTTDirectionGetter.from_shcoeff,
            fit.shm_coeff,
            90,
            unit_octahedron,
            basis_type="tournier07",
            probe_length=0,
        )

        # Check probe_radius ValueError
        npt.assert_raises(
            ValueError,
            PTTDirectionGetter.from_shcoeff,
            fit.shm_coeff,
            90,
            unit_octahedron,
            basis_type="tournier07",
            probe_radius=-1,
        )

        # Check probe_quality ValueError
        npt.assert_raises(
            ValueError,
            PTTDirectionGetter.from_shcoeff,
            fit.shm_coeff,
            90,
            unit_octahedron,
            basis_type="tournier07",
            probe_quality=1,
        )

        # Check probe_length ValueError
        npt.assert_raises(
            ValueError,
            PTTDirectionGetter.from_shcoeff,
            fit.shm_coeff,
            90,
            unit_octahedron,
            basis_type="tournier07",
            probe_count=0,
        )



"""

Visualization tools for 2D projections of 3D functions on the sphere, such as
ODFs.

"""

import numpy as np
import scipy.interpolate as interp

import dipy.core.geometry as geo
from dipy.testing.decorators import doctest_skip_parser, warning_for_keywords
from dipy.utils.optpkg import optional_package

matplotlib, has_mpl, setup_module = optional_package("matplotlib")
plt, _, _ = optional_package("matplotlib.pyplot")
bm, has_basemap, _ = optional_package("mpl_toolkits.basemap")


@warning_for_keywords()
@doctest_skip_parser
def sph_project(
    vertices,
    val,
    *,
    ax=None,
    vmin=None,
    vmax=None,
    cmap=None,
    cbar=True,
    tri=False,
    boundary=False,
    **basemap_args,
):
    """Draw a signal on a 2D projection of the sphere.

    Parameters
    ----------
    vertices : (N,3) ndarray
        Unit vector points of the sphere
    val : (N) ndarray
        Function values.
    ax : mpl axis, optional
        If specified, draw onto this existing axis instead.
    vmin : float, optional
       Minimum value to cut the z.
    vmax : float, optional
       Minimum value to cut the z.
    cmap : matplotlib.colors.Colormap, optional
        Colormap.
    cbar : bool, optional
        Whether to add the color-bar to the figure.
    tri : bool, optional
        Whether to display the plot triangulated as a pseudo-color plot.
    boundary : bool, optional
        Whether to draw the boundary around the projection in a black line.

    Returns
    -------
    ax : axis
        Matplotlib figure axis

    Examples
    --------
    >>> from dipy.data import default_sphere
    >>> verts = default_sphere.vertices
    >>> _ax = sph_project(verts.T, np.random.rand(len(verts.T))) # skip if not has_basemap
    """  # noqa: E501
    if ax is None:
        fig, ax = plt.subplots(1)
    else:
        fig = plt.subplots(1)
        fig.axes.append(ax)

    if cmap is None:
        cmap = matplotlib.cm.hot

    basemap_args.setdefault("projection", "ortho")
    basemap_args.setdefault("lat_0", 0)
    basemap_args.setdefault("lon_0", 0)
    basemap_args.setdefault("resolution", "c")

    m = bm.Basemap(**basemap_args)
    if boundary:
        m.drawmapboundary()

    # Rotate the coordinate system so that you are looking from the north pole:
    verts_rot = np.array(np.dot(np.array([[0, 0, -1], [0, 1, 0], [1, 0, 0]]), vertices))

    # To get the orthographic projection, when the first coordinate is
    # positive:
    neg_idx = np.where(verts_rot[0] > 0)

    # rotate the entire b-vector around to point in the other direction:
    verts_rot[:, neg_idx] *= -1

    _, theta, phi = geo.cart2sphere(verts_rot[0], verts_rot[1], verts_rot[2])
    lat, lon = geo.sph2latlon(theta, phi)
    x, y = m(lon, lat)

    my_min = np.nanmin(val)
    if vmin is not None:
        my_min = vmin

    my_max = np.nanmax(val)
    if vmax is not None:
        my_max = vmax

    if tri:
        m.pcolor(x, y, val, vmin=my_min, vmax=my_max, tri=True, cmap=cmap)

    else:
        cmap_data = cmap._segmentdata
        red_interp, blue_interp, green_interp = (
            interp.interp1d(
                np.array(cmap_data[gun])[:, 0], np.array(cmap_data[gun])[:, 1]
            )
            for gun in ["red", "blue", "green"]
        )

        r = (val - my_min) / float(my_max - my_min)

        # Enforce the maximum and minimum boundaries, if there are values
        # outside those boundaries:
        r[r < 0] = 0
        r[r > 1] = 1

        for this_x, this_y, this_r in zip(x, y, r):
            red = red_interp(this_r)
            blue = blue_interp(this_r)
            green = green_interp(this_r)
            m.plot(this_x, this_y, "o", c=[red.item(), green.item(), blue.item()])

    if cbar:
        mappable = matplotlib.cm.ScalarMappable(cmap=cmap)
        mappable.set_array([my_min, my_max])
        # setup colorbar axes instance.
        pos = ax.get_position()
        ell, b, w, h = pos.bounds
        # setup colorbar axes
        cax = fig.add_axes([ell + w + 0.075, b, 0.05, h], frameon=False)
        fig.colorbar(mappable, cax=cax)  # draw colorbar

    return ax



"""
==============================
Introduction to Basic Tracking
==============================

Local fiber tracking is an approach used to model white matter fibers by
creating streamlines from local directional information. The idea is as
follows: if the local directionality of a tract/pathway segment is known, one
can integrate along those directions to build a complete representation of that
structure. Local fiber tracking is widely used in the field of diffusion MRI
because it is simple and robust.

In order to perform local fiber tracking, three things are needed:

1. A method for getting directions from a diffusion dataset.
2. A method for identifying when the tracking must stop.
3. A set of seeds from which to begin tracking.

This example shows how to combine the 3 parts described above
to create a tractography reconstruction from a diffusion data set.

Let's begin by importing the necessary modules.
"""

import matplotlib.pyplot as plt

from dipy.core.gradients import gradient_table
from dipy.data import default_sphere, get_fnames
from dipy.direction import peaks_from_model
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti, load_nifti_data
from dipy.io.stateful_tractogram import Space, StatefulTractogram
from dipy.io.streamline import save_trk
from dipy.reconst.shm import CsaOdfModel
from dipy.tracking import utils
from dipy.tracking.stopping_criterion import ThresholdStoppingCriterion
from dipy.tracking.streamline import Streamlines
from dipy.tracking.tracker import eudx_tracking
from dipy.viz import actor, colormap, has_fury, window

###############################################################################
# Now, let's load an HARDI dataset from Stanford. If you have
# not already downloaded this data set, the first time you run this example you
# will need to be connected to the internet and this dataset will be downloaded
# to your computer.

# Enables/disables interactive visualization
interactive = False

hardi_fname, hardi_bval_fname, hardi_bvec_fname = get_fnames(name="stanford_hardi")
label_fname = get_fnames(name="stanford_labels")

data, affine, hardi_img = load_nifti(hardi_fname, return_img=True)
labels = load_nifti_data(label_fname)
bvals, bvecs = read_bvals_bvecs(hardi_bval_fname, hardi_bvec_fname)
gtab = gradient_table(bvals, bvecs=bvecs)

###############################################################################
# This dataset provides a label map in which all white matter tissues are
# labeled either 1 or 2. Let's create a white matter mask to restrict tracking
# to the white matter.

white_matter = (labels == 1) | (labels == 2)

###############################################################################
# Step 1: Getting directions from a diffusion dataset
# ---------------------------------------------------
#
# The first thing we need to begin fiber tracking is a way of getting
# directions from this diffusion data set. In order to do that, we can fit the
# data to a Constant Solid Angle ODF Model. This model will estimate the
# Orientation Distribution Function (ODF) at each voxel. The ODF is the
# distribution of water diffusion as a function of direction. The peaks of an
# ODF are good estimates for the orientation of tract segments at a point in
# the image. Here, we use ``peaks_from_model`` to fit the data and calculate
# the fiber directions in all voxels of the white matter.

csa_model = CsaOdfModel(gtab, sh_order_max=6)
csa_peaks = peaks_from_model(
    csa_model,
    data,
    default_sphere,
    relative_peak_threshold=0.8,
    min_separation_angle=45,
    mask=white_matter,
)

###############################################################################
# For quality assurance we can also visualize a slice from the direction field
# which we will use as the basis to perform the tracking. The visualization
# will be done using the ``fury`` python package

if has_fury:
    scene = window.Scene()
    scene.add(
        actor.peak_slicer(
            csa_peaks.peak_dirs, peaks_values=csa_peaks.peak_values, colors=None
        )
    )

    window.record(scene=scene, out_path="csa_direction_field.png", size=(900, 900))

    if interactive:
        window.show(scene, size=(800, 800))

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Direction Field (peaks)
#
#
# Step 2: Identifying when the tracking must stop
# -----------------------------------------------
# Next we need some way of restricting the fiber tracking to areas with good
# directionality information. We've already created the white matter mask,
# but we can go a step further and restrict fiber tracking to those areas where
# the ODF shows significant restricted diffusion by thresholding on
# the generalized fractional anisotropy (GFA).

stopping_criterion = ThresholdStoppingCriterion(csa_peaks.gfa, 0.25)

###############################################################################
# Again, for quality assurance, we can also visualize a slice of the GFA and
# the resulting tracking mask.

sli = csa_peaks.gfa.shape[2] // 2
plt.figure("GFA")
plt.subplot(1, 2, 1).set_axis_off()
plt.imshow(csa_peaks.gfa[:, :, sli].T, cmap="gray", origin="lower")

plt.subplot(1, 2, 2).set_axis_off()
plt.imshow((csa_peaks.gfa[:, :, sli] > 0.25).T, cmap="gray", origin="lower")

plt.savefig("gfa_tracking_mask.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# An example of a tracking mask derived from the generalized fractional
# anisotropy (GFA).
#
#
#
# Step 3: Defining a set of seeds from which to begin tracking
# ------------------------------------------------------------
# Before we can begin tracking, we need to specify where to "seed" (begin)
# the fiber tracking. Generally, the seeds chosen will depend on the pathways
# one is interested in modeling. In this example, we'll use a
# $2 \times 2 \times 2$ grid of seeds per voxel, in a sagittal slice of the
# corpus callosum. Tracking from this region will give us a model of the
# corpus callosum tract. This slice has label value ``2`` in the label's image.

seed_mask = labels == 2
seeds = utils.seeds_from_mask(seed_mask, affine, density=[2, 2, 2])

###############################################################################
# Finally, we can bring it all together using ``eudx_tracking``, using
# the EuDX algorithm :footcite:p:`Garyfallidis2012b`. ``EuDX`` is a fast
# algorithm that we use here to generate streamlines. This algorithm is what is
# used here and the default option when providing the output of peaks directly
# in eudx_tracking.

# Initialization of eudx_tracking. The computation happens in the next step.
streamlines_generator = eudx_tracking(
    seeds, stopping_criterion, affine, step_size=0.5, pam=csa_peaks
)
# Generate streamlines object
streamlines = Streamlines(streamlines_generator)

###############################################################################
# We will then display the resulting streamlines using the ``fury``
# python package.

if has_fury:
    # Prepare the display objects.
    color = colormap.line_colors(streamlines)

    streamlines_actor = actor.line(
        streamlines, colors=colormap.line_colors(streamlines)
    )

    # Create the 3D display.
    scene = window.Scene()
    scene.add(streamlines_actor)

    # Save still images for this static example. Or for interactivity use
    window.record(scene=scene, out_path="tractogram_EuDX.png", size=(800, 800))
    if interactive:
        window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Corpus Callosum using EuDx
#
#
# We've created a deterministic set of streamlines using the EuDX algorithm.
# This is so called deterministic because if you repeat the fiber tracking
# (keeping all the inputs the same) you will get exactly the same set of
# streamlines. We can save the streamlines as a Trackvis file so it can be
# loaded into other software for visualization or further analysis.

sft = StatefulTractogram(streamlines, hardi_img, Space.RASMM)
save_trk(sft, "tractogram_EuDX.trk", streamlines)

###############################################################################
# References
# ----------
#
# .. footbibliography::
#



import os
from os.path import join as pjoin
from tempfile import TemporaryDirectory
import warnings

import numpy as np
import numpy.testing as npt

from dipy.core.gradients import gradient_table
from dipy.core.subdivide_octahedron import create_unit_sphere
from dipy.data import default_sphere, get_fnames
from dipy.direction.peaks import PeaksAndMetrics
from dipy.io.image import load_nifti
from dipy.io.peaks import (
    load_pam,
    load_peaks,
    niftis_to_pam,
    pam_to_niftis,
    save_pam,
    save_peaks,
    tensor_to_pam,
)
import dipy.reconst.dti as dti
from dipy.testing.decorators import set_random_number_generator


def generate_default_pam(rng):
    pam = PeaksAndMetrics()
    pam.affine = np.eye(4)
    pam.peak_dirs = rng.random((10, 10, 10, 5, 3))
    pam.peak_values = np.zeros((10, 10, 10, 5))
    pam.peak_indices = np.zeros((10, 10, 10, 5))
    pam.shm_coeff = np.zeros((10, 10, 10, 45))
    pam.sphere = default_sphere
    pam.B = np.zeros((45, default_sphere.vertices.shape[0]))
    pam.total_weight = 0.5
    pam.ang_thr = 60
    pam.gfa = np.zeros((10, 10, 10))
    pam.qa = np.zeros((10, 10, 10, 5))
    pam.odf = np.zeros((10, 10, 10, default_sphere.vertices.shape[0]))
    return pam


@set_random_number_generator()
def test_io_peaks(rng):
    with TemporaryDirectory() as tmpdir:
        fname = pjoin(tmpdir, "test.pam5")

        pam = generate_default_pam(rng)
        save_pam(fname, pam)
        pam2 = load_pam(fname, verbose=False)
        npt.assert_array_equal(pam.peak_dirs, pam2.peak_dirs)

        pam2.affine = None

        fname2 = pjoin(tmpdir, "test2.pam5")
        save_pam(fname2, pam2, affine=np.eye(4))
        pam2_res = load_pam(fname2, verbose=True)
        npt.assert_array_equal(pam.peak_dirs, pam2_res.peak_dirs)

        pam3 = load_pam(fname2, verbose=False)

        for attr in [
            "peak_dirs",
            "peak_values",
            "peak_indices",
            "gfa",
            "qa",
            "shm_coeff",
            "B",
            "odf",
        ]:
            npt.assert_array_equal(getattr(pam3, attr), getattr(pam, attr))

        npt.assert_equal(pam3.total_weight, pam.total_weight)
        npt.assert_equal(pam3.ang_thr, pam.ang_thr)
        npt.assert_array_almost_equal(pam3.sphere.vertices, pam.sphere.vertices)

        fname3 = pjoin(tmpdir, "test3.pam5")
        pam4 = PeaksAndMetrics()
        npt.assert_raises((ValueError, AttributeError), save_pam, fname3, pam4)

        fname4 = pjoin(tmpdir, "test4.pam5")
        del pam.affine
        save_pam(fname4, pam, affine=None)

        fname5 = pjoin(tmpdir, "test5.pkm")
        npt.assert_raises(IOError, save_pam, fname5, pam)

        pam.affine = np.eye(4)
        fname6 = pjoin(tmpdir, "test6.pam5")
        save_pam(fname6, pam, verbose=True)

        del pam.shm_coeff
        save_pam(pjoin(tmpdir, fname6), pam, verbose=False)

        pam.shm_coeff = np.zeros((10, 10, 10, 45))
        del pam.odf
        save_pam(fname6, pam)
        pam_tmp = load_pam(fname6, verbose=True)
        npt.assert_equal(pam_tmp.odf, None)

        fname7 = pjoin(tmpdir, "test7.paw")
        npt.assert_raises(OSError, load_pam, fname7)

        del pam.shm_coeff
        save_pam(fname6, pam, verbose=True)

        fname_shm = pjoin(tmpdir, "shm.nii.gz")
        fname_dirs = pjoin(tmpdir, "peaks_dirs.nii.gz")
        fname_values = pjoin(tmpdir, "peaks_values.nii.gz")
        fname_indices = pjoin(tmpdir, "peaks_indices.nii.gz")
        fname_gfa = pjoin(tmpdir, "gfa.nii.gz")
        fname_sphere = pjoin(tmpdir, "sphere.txt")
        fname_b = pjoin(tmpdir, "B.nii.gz")
        fname_qa = pjoin(tmpdir, "qa.nii.gz")

        pam.shm_coeff = np.ones((10, 10, 10, 45))
        pam_to_niftis(
            pam,
            fname_shm=fname_shm,
            fname_peaks_dir=fname_dirs,
            fname_peaks_values=fname_values,
            fname_peaks_indices=fname_indices,
            fname_gfa=fname_gfa,
            fname_sphere=fname_sphere,
            fname_b=fname_b,
            fname_qa=fname_qa,
            reshape_dirs=False,
        )

        for name in [
            "shm.nii.gz",
            "peaks_dirs.nii.gz",
            "peaks_values.nii.gz",
            "gfa.nii.gz",
            "peaks_indices.nii.gz",
            "shm.nii.gz",
        ]:
            npt.assert_(
                os.path.isfile(pjoin(tmpdir, name)),
                "{} file does not exist".format(pjoin(tmpdir, name)),
            )


@set_random_number_generator()
def test_io_save_pam_error(rng):
    with TemporaryDirectory() as tmpdir:
        fname = "test.pam5"

        pam = PeaksAndMetrics()

        npt.assert_raises(IOError, save_pam, pjoin(tmpdir, "test.pam"), pam)
        npt.assert_raises(
            (ValueError, AttributeError), save_pam, pjoin(tmpdir, fname), pam
        )

        pam.affine = np.eye(4)
        pam.peak_dirs = rng.random((10, 10, 10, 5, 3))
        pam.peak_values = np.zeros((10, 10, 10, 5))
        pam.peak_indices = np.zeros((10, 10, 10, 5))
        pam.shm_coeff = np.zeros((10, 10, 10, 45))
        pam.sphere = default_sphere
        pam.B = np.zeros((45, default_sphere.vertices.shape[0]))
        pam.total_weight = 0.5
        pam.ang_thr = 60
        pam.gfa = np.zeros((10, 10, 10))
        pam.qa = np.zeros((10, 10, 10, 5))
        pam.odf = np.zeros((10, 10, 10, default_sphere.vertices.shape[0]))


def test_io_niftis_to_pam():
    with TemporaryDirectory() as tmpdir:
        pam = niftis_to_pam(
            affine=np.eye(4),
            peak_dirs=np.random.rand(10, 10, 10, 5, 3),
            peak_values=np.zeros((10, 10, 10, 5)),
            peak_indices=np.zeros((10, 10, 10, 5)),
            shm_coeff=np.zeros((10, 10, 10, 45)),
            sphere=default_sphere,
            gfa=np.zeros((10, 10, 10)),
            B=np.zeros((45, default_sphere.vertices.shape[0])),
            qa=np.zeros((10, 10, 10, 5)),
            odf=np.zeros((10, 10, 10, default_sphere.vertices.shape[0])),
            total_weight=0.5,
            ang_thr=60,
            pam_file=pjoin(tmpdir, "test15.pam5"),
        )

        npt.assert_equal(pam.peak_dirs.shape, (10, 10, 10, 5, 3))
        npt.assert_(os.path.isfile(pjoin(tmpdir, "test15.pam5")))


def test_tensor_to_pam():
    fdata, fbval, fbvec = get_fnames(name="small_25")
    gtab = gradient_table(fbval, bvecs=fbvec)
    data, affine = load_nifti(fdata)
    dm = dti.TensorModel(gtab)
    df = dm.fit(data)
    df.evals[0, 0, 0] = np.array([0, 0, 0])
    sphere = create_unit_sphere(recursion_level=4)
    odf = df.odf(sphere)

    with TemporaryDirectory() as tmpdir:
        fname = "test_tt.pam5"
        pam = tensor_to_pam(
            evals=df.evals,
            evecs=df.evecs,
            affine=affine,
            sphere=sphere,
            odf=odf,
            pam_file=pjoin(tmpdir, fname),
        )
        npt.assert_(os.path.isfile(pjoin(tmpdir, fname)))
        save_pam(pjoin(tmpdir, "test_tt_2.pam5"), pam)
        pam2 = load_pam(pjoin(tmpdir, "test_tt_2.pam5"))

        npt.assert_array_equal(pam.peak_values, pam2.peak_values)
        npt.assert_array_equal(pam.peak_dirs, pam2.peak_dirs)
        npt.assert_array_almost_equal(pam.peak_indices, pam2.peak_indices)
        del pam


@set_random_number_generator()
def test_io_peaks_deprecated(rng):
    with TemporaryDirectory() as tmpdir:
        with warnings.catch_warnings(record=True) as cw:
            warnings.simplefilter("always", DeprecationWarning)
            fname = pjoin(tmpdir, "test_tt.pam5")
            pam = generate_default_pam(rng)
            save_peaks(fname, pam)
            pam2 = load_peaks(fname, verbose=True)
            npt.assert_array_equal(pam.peak_dirs, pam2.peak_dirs)
            npt.assert_equal(len(cw), 2)
            npt.assert_(issubclass(cw[0].category, DeprecationWarning))
            npt.assert_(issubclass(cw[1].category, DeprecationWarning))



"""Testing reconstruction utilities."""

import numpy as np
from numpy.testing import assert_array_equal, assert_equal

from dipy.reconst.recspeed import adj_to_countarrs, argmax_from_countarrs


def test_adj_countarrs():
    adj = [[0, 1, 2], [2, 3], [4, 5, 6, 7]]
    counts, inds = adj_to_countarrs(adj)
    assert_array_equal(counts, [3, 2, 4])
    assert_equal(counts.dtype.type, np.uint32)
    assert_array_equal(inds, [0, 1, 2, 2, 3, 4, 5, 6, 7])
    assert_equal(inds.dtype.type, np.uint32)


def test_argmax_from_countarrs():
    # basic case
    vals = np.arange(10, dtype=float)
    vertinds = np.arange(10, dtype=np.uint32)
    adj_counts = np.ones((10,), dtype=np.uint32)
    adj_inds_raw = np.arange(10, dtype=np.uint32)[::-1]
    # when contiguous - OK
    adj_inds = adj_inds_raw.copy()
    argmax_from_countarrs(vals, vertinds, adj_counts, adj_inds)
    # yield assert_array_equal(inds, [5, 6, 7, 8, 9])
    # test for errors - first - not contiguous
    #
    # The tests below cause odd errors and segfaults with numpy SVN
    # vintage June 2010 (sometime after 1.4.0 release) - see
    # http://groups.google.com/group/cython-users/browse_thread/thread/624c696293b7fe44?pli=1
    """
    yield assert_raises(ValueError,
                        argmax_from_countarrs,
                        vals,
                        vertinds,
                        adj_counts,
                        adj_inds_raw)
    # too few vertices
    yield assert_raises(ValueError,
                        argmax_from_countarrs,
                        vals,
                        vertinds[:-1],
                        adj_counts,
                        adj_inds)
    # adj_inds too short
    yield assert_raises(IndexError,
                        argmax_from_countarrs,
                        vals,
                        vertinds,
                        adj_counts,
                        adj_inds[:-1])
    # vals too short
    yield assert_raises(IndexError,
                        argmax_from_countarrs,
                        vals[:-1],
                        vertinds,
                        adj_counts,
                        adj_inds)
                        """



"""
====================================================
Extracting AFQ tract profiles from segmented bundles
====================================================

In this example, we will extract the values of a statistic from a
volume, using the coordinates along the length of a bundle. These are called
`tract profiles`

One of the challenges of extracting tract profiles is that some of the
streamlines in a bundle may diverge significantly from the bundle in some
locations. To overcome this challenge, we will use a strategy similar to that
described in :footcite:p:`Yeatman2012`: We will weight the contribution of each
streamline to the bundle profile based on how far this streamline is from the
mean trajectory of the bundle at that location.

"""

import os.path as op

import matplotlib.pyplot as plt
import numpy as np

from dipy.data import fetch_hbn, get_two_hcp842_bundles
from dipy.io.image import load_nifti
from dipy.io.streamline import load_trk
from dipy.segment.clustering import QuickBundles
from dipy.segment.featurespeed import ResampleFeature
from dipy.segment.metricspeed import AveragePointwiseEuclideanMetric
import dipy.stats.analysis as dsa
import dipy.tracking.streamline as dts

###############################################################################
# To demonstrate this, we will use data from the Healthy Brain Network (HBN)
# study :footcite:p:`Alexander2017` that has already been processed
# :footcite:p:`RichieHalford2022`. For this demonstration, we will use only the
# left arcuate fasciculus (ARC) and the left corticospinal tract (CST) from the
# subject NDARAA948VFH.

subject = "NDARAA948VFH"
session = "HBNsiteRU"

fdict, path = fetch_hbn([subject], include_afq=True)

afq_path = op.join(path, "derivatives", "afq", f"sub-{subject}", f"ses-{session}")

###############################################################################
# We can use the `dipy.io` API to read in the bundles from file.
# `load_trk` returns both the streamlines, as well as header information, and
# the `streamlines` attribute will give us access to the sequence of arrays
# that contain the streamline coordinates.

cst_l_file = op.join(
    afq_path,
    "clean_bundles",
    f"sub-{subject}_ses-{session}_acq-64dir_space-T1w_desc-preproc_dwi_space"
    "-RASMM_model-CSD_desc-prob-afq-CST_L_tractography.trk",
)

arc_l_file = op.join(
    afq_path,
    "clean_bundles",
    f"sub-{subject}_ses-{session}_acq-64dir_space-T1w_desc-preproc_dwi_space"
    "-RASMM_model-CSD_desc-prob-afq-ARC_L_tractography.trk",
)

cst_l = load_trk(cst_l_file, reference="same", bbox_valid_check=False).streamlines
arc_l = load_trk(arc_l_file, reference="same", bbox_valid_check=False).streamlines

###############################################################################
# In the next step, we need to make sure that all the streamlines in each
# bundle are oriented the same way. For example, for the CST, we want to make
# sure that all the bundles have their cortical termination at one end of the
# streamline. This is so that when we later extract values from a volume,
# we will not have different streamlines going in opposite directions.
#
# To orient all the streamlines in each bundle, we will create standard
# streamlines, by finding the centroids of the left ARC and CST bundle models.
#
# The advantage of using the model bundles is that we can use the same
# standard for different subjects, which means that we'll get the same
# orientation of the streamlines in all subjects.

model_arc_l_file, model_cst_l_file = get_two_hcp842_bundles()

model_arc_l = load_trk(
    model_arc_l_file, reference="same", bbox_valid_check=False
).streamlines
model_cst_l = load_trk(
    model_cst_l_file, reference="same", bbox_valid_check=False
).streamlines


feature = ResampleFeature(nb_points=100)
metric = AveragePointwiseEuclideanMetric(feature)

###############################################################################
# Since we are going to include all of the streamlines in the single cluster
# from the streamlines, we set the threshold to `np.inf`. We pull out the
# centroid as the standard using QuickBundles :footcite:p:`Garyfallidis2012a`.

qb = QuickBundles(threshold=np.inf, metric=metric)

cluster_cst_l = qb.cluster(model_cst_l)
standard_cst_l = cluster_cst_l.centroids[0]

cluster_af_l = qb.cluster(model_arc_l)
standard_af_l = cluster_af_l.centroids[0]

###############################################################################
# We use the centroid streamline for each atlas bundle as the standard to
# orient all of the streamlines in each bundle from the individual subject.
# Here, the affine used is the one from the transform between the atlas and
# individual tractogram. This is so that the orienting is done relative to the
# space of the individual, and not relative to the atlas space.

oriented_cst_l = dts.orient_by_streamline(cst_l, standard_cst_l)
oriented_arc_l = dts.orient_by_streamline(arc_l, standard_af_l)

###############################################################################
# Tract profiles are created from a scalar property of the volume. Here, we
# read volumetric data from an image corresponding to the FA calculated in
# this subject with the diffusion tensor imaging (DTI) model.

fa, fa_affine = load_nifti(
    op.join(
        afq_path,
        f"sub-{subject}_ses-{session}_acq-64dir_space-T1w_desc"
        "-preproc_dwi_model-DTI_FA.nii.gz",
    )
)

###############################################################################
# As mentioned at the outset, we would like to downweight the streamlines that
# are far from the core trajectory of the tracts. We calculate
# weights for each bundle:

w_cst_l = dsa.gaussian_weights(oriented_cst_l)
w_arc_l = dsa.gaussian_weights(oriented_arc_l)

###############################################################################
# And then use the weights to calculate the tract profiles for each bundle

profile_cst_l = dsa.afq_profile(fa, oriented_cst_l, affine=fa_affine, weights=w_cst_l)

profile_af_l = dsa.afq_profile(fa, oriented_arc_l, affine=fa_affine, weights=w_arc_l)

fig, (ax1, ax2) = plt.subplots(1, 2)

ax1.plot(profile_cst_l)
ax1.set_ylabel("Fractional anisotropy")
ax1.set_xlabel("Node along CST")
ax2.plot(profile_af_l)
ax2.set_xlabel("Node along ARC")
fig.savefig("tract_profiles")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Bundle profiles for the fractional anisotropy in the left CST (left) and
# left AF (right).
#
#
# References
# ----------
#
# .. footbibliography::
#



"""
=====================
Gradients and Spheres
=====================

This example shows how you can create gradient tables and sphere objects using
DIPY_.

Usually, as we saw in
:ref:`sphx_glr_examples_built_quick_start_quick_start.py`,
you load your b-values and b-vectors from disk and then you can create your own
gradient table. But this time let's say that you are an MR physicist and you
want to design a new gradient scheme or you are a scientist who wants to
simulate many different gradient schemes.

Now let's assume that you are interested in creating a multi-shell
acquisition with 2-shells, one at b=1000 $s/mm^2$ and one at b=2500 $s/mm^2$.
For both shells let's say that we want a specific number of gradients (64) and
we want to have the points on the sphere evenly distributed.

This is possible using the ``disperse_charges`` which is an implementation of
electrostatic repulsion :footcite:t:`Jones1999` .

Let's start by importing the necessary modules.
"""

import numpy as np

from dipy.core.gradients import gradient_table
from dipy.core.sphere import HemiSphere, Sphere, disperse_charges
from dipy.viz import actor, window

###############################################################################
# We can first create some random points on a ``HemiSphere`` using spherical
# polar coordinates.

rng = np.random.default_rng()
n_pts = 64
theta = np.pi * rng.random(n_pts)
phi = 2 * np.pi * rng.random(n_pts)
hsph_initial = HemiSphere(theta=theta, phi=phi)

###############################################################################
# Next, we call ``disperse_charges`` which will iteratively move the points so
# that the electrostatic potential energy is minimized.

hsph_updated, potential = disperse_charges(hsph_initial, 5000)

###############################################################################
# In ``hsph_updated`` we have the updated ``HemiSphere`` with the points nicely
# distributed on the hemisphere. Let's visualize them.

# Enables/disables interactive visualization
interactive = False

scene = window.Scene()
scene.SetBackground(1, 1, 1)

scene.add(actor.point(hsph_initial.vertices, window.colors.red, point_radius=0.05))
scene.add(actor.point(hsph_updated.vertices, window.colors.green, point_radius=0.05))

window.record(scene=scene, out_path="initial_vs_updated.png", size=(300, 300))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Illustration of electrostatic repulsion of red points which become
# green points.
#
#
# We can also create a sphere from the hemisphere and show it in the
# following way.

sph = Sphere(xyz=np.vstack((hsph_updated.vertices, -hsph_updated.vertices)))

scene.clear()
scene.add(actor.point(sph.vertices, window.colors.green, point_radius=0.05))

window.record(scene=scene, out_path="full_sphere.png", size=(300, 300))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Full sphere.
#
#
# It is time to create the Gradients. For this purpose we will use the
# function ``gradient_table`` and fill it with the ``hsph_updated`` vectors
# that we created above.

vertices = hsph_updated.vertices
values = np.ones(vertices.shape[0])

###############################################################################
# We need two stacks of ``vertices``, one for every shell, and we need two sets
# of b-values, one at 1000 $s/mm^2$, and one at 2500 $s/mm^2$, as we discussed
# previously.

bvecs = np.vstack((vertices, vertices))
bvals = np.hstack((1000 * values, 2500 * values))

###############################################################################
# We can also add some b0s. Let's add one at the beginning and one at the end.

bvecs = np.insert(bvecs, (0, bvecs.shape[0]), np.array([0, 0, 0]), axis=0)
bvals = np.insert(bvals, (0, bvals.shape[0]), 0)

print(bvals)
print(bvecs)

###############################################################################
# Both b-values and b-vectors look correct. Let's now create the
# ``GradientTable``.

gtab = gradient_table(bvals, bvecs=bvecs)

scene.clear()

###############################################################################
# We can also visualize the gradients. Let's color the first shell blue and
# the second shell cyan.

colors_b1000 = window.colors.blue * np.ones(vertices.shape)
colors_b2500 = window.colors.cyan * np.ones(vertices.shape)
colors = np.vstack((colors_b1000, colors_b2500))
colors = np.insert(colors, (0, colors.shape[0]), np.array([0, 0, 0]), axis=0)
colors = np.ascontiguousarray(colors)

scene.add(actor.point(gtab.gradients, colors, point_radius=100))

window.record(scene=scene, out_path="gradients.png", size=(300, 300))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Diffusion gradients.
#
#
# References
# ----------
#
# .. footbibliography::



# cython: boundscheck=False
# cython: initializedcheck=False
# cython: wraparound=False
# cython: Nonecheck=False

"""Track propagation performance functions."""

# cython: profile=True
# cython: embedsignature=True

cimport cython
from cython.parallel import prange

import numpy as np
cimport numpy as cnp

from dipy.core.interpolation cimport _trilinear_interpolation_iso, offset
from dipy.direction.pmf cimport PmfGen
from dipy.utils.fast_numpy cimport (
    copy_point,
    cross,
    cumsum,
    norm,
    normalize,
    random_float,
    random_perpendicular_vector,
    random_point_within_circle,
    RNGState,
    where_to_insert,
)
from dipy.tracking.tractogen cimport prepare_pmf
from dipy.tracking.tracker_parameters cimport (TrackerParameters, TrackerStatus,
                                               SUCCESS, FAIL)

from libc.stdlib cimport malloc, free
from libc.math cimport M_PI, pow, sin, cos, fabs
from libc.stdio cimport printf

cdef extern from "dpy_math.h" nogil:
    double floor(double x)
    float acos(float x )
    double sqrt(double x)
    double DPY_PI


DEF PEAK_NO=5

# initialize numpy runtime
cnp.import_array()


def ndarray_offset(cnp.ndarray[cnp.npy_intp, ndim=1] indices,
                   cnp.ndarray[cnp.npy_intp, ndim=1] strides,
                   int lenind,
                   int typesize):
    """ Find offset in an N-dimensional ndarray using strides

    Parameters
    ----------
    indices : array, npy_intp shape (N,)
        Indices of the array which we want to find the offset.
    strides : array, shape (N,)
        Strides of array.
    lenind : int
        len of the `indices` array.
    typesize : int
        Number of bytes for data type e.g. if 8 for double, 4 for int32

    Returns
    -------
    offset : integer
        Index position in flattened array

    Examples
    --------
    >>> import numpy as np
    >>> from dipy.tracking.propspeed import ndarray_offset
    >>> I=np.array([1,1])
    >>> A=np.array([[1,0,0],[0,2,0],[0,0,3]])
    >>> S=np.array(A.strides)
    >>> ndarray_offset(I,S,2,A.dtype.itemsize)
    4
    >>> A.ravel()[4]==A[1,1]
    True
    """
    if not cnp.PyArray_CHKFLAGS(indices, cnp.NPY_ARRAY_C_CONTIGUOUS):
        raise ValueError("indices is not C contiguous")
    if not cnp.PyArray_CHKFLAGS(strides, cnp.NPY_ARRAY_C_CONTIGUOUS):
        raise ValueError("strides is not C contiguous")
    return offset(<cnp.npy_intp*> cnp.PyArray_DATA(indices),
                  <cnp.npy_intp*> cnp.PyArray_DATA(strides),
                  lenind,
                  typesize)


cdef cnp.npy_intp _nearest_direction(double* dx,
                                     double* qa,
                                     double *ind,
                                     cnp.npy_intp peaks,
                                     double *odf_vertices,
                                     double qa_thr, double ang_thr,
                                     double *direction) noexcept nogil:
    """ Give the nearest direction to a point, checking threshold and angle

    Parameters
    ----------
    dx : double array shape (3,)
        Moving direction of the current tracking.
    qa : double array shape (Np,)
        Quantitative anisotropy matrix, where ``Np`` is the number of peaks.
    ind : array, float64 shape(x, y, z, Np)
        Index of the track orientation.
    peaks : npy_intp
    odf_vertices : double array shape (N, 3)
        Sampling directions on the sphere.
    qa_thr : float
        Threshold for QA, we want everything higher than this threshold.
    ang_thr : float
        Angle threshold, we only select fiber orientation within this range.
    direction : double array shape (3,)
        The fiber orientation to be considered in the interpolation.  The array
        gets modified in-place.

    Returns
    -------
    delta : bool
        Delta function: if 1 we give it weighting, if it is 0 we don't give any
        weighting.
    """
    cdef:
        double max_dot = 0
        double angl,curr_dot
        double odfv[3]
        cnp.npy_intp i, j, max_doti = 0

    # calculate the cos with radians
    angl = cos((DPY_PI * ang_thr) / 180.)
    # if the maximum peak is lower than the threshold then there is no point
    # continuing tracking
    if qa[0] <= qa_thr:
        return 0
    # for all peaks find the minimum angle between odf_vertices and dx
    for i from 0 <= i < peaks:
        # if the current peak is smaller than the threshold then jump out
        if qa[i] <= qa_thr:
            break
        # copy odf_vertices
        for j from 0 <= j < 3:
            odfv[j]=odf_vertices[3 * <cnp.npy_intp>ind[i] + j]
        # calculate the absolute dot product between dx and odf_vertices
        curr_dot = dx[0] * odfv[0] + dx[1] * odfv[1] + dx[2] * odfv[2]
        if curr_dot < 0: #abs check
            curr_dot = -curr_dot
        # maximum dot means minimum angle
        # store the maximum dot and the corresponding index from the
        # neighboring voxel in maxdoti
        if curr_dot > max_dot:
            max_dot=curr_dot
            max_doti = i
    # if maxdot smaller than our angular *dot* threshold stop tracking
    if max_dot < angl:
        return 0
    # copy the odf_vertices for the voxel qa indices which have the smaller
    # angle
    for j from 0 <= j < 3:
        odfv[j] = odf_vertices[3 * <cnp.npy_intp>ind[max_doti] + j]
    # if the dot product is negative then return the opposite direction
    # otherwise return the same direction
    if dx[0] * odfv[0] + dx[1] * odfv[1] + dx[2] * odfv[2] < 0:
        for j from 0 <= j < 3:
            direction[j] = -odf_vertices[3 * <cnp.npy_intp>ind[max_doti] + j]
        return 1
    for j from 0 <= j < 3:
        direction[j]= odf_vertices[3 * <cnp.npy_intp>ind[max_doti] + j]
    return 1


@cython.cdivision(True)
cdef cnp.npy_intp _propagation_direction(double *point,
                                         double* dx,
                                         double* qa,
                                         double *ind,
                                         double *odf_vertices,
                                         double qa_thr,
                                         double ang_thr,
                                         cnp.npy_intp *qa_shape,
                                         cnp.npy_intp* strides,
                                         double *direction,
                                         double total_weight) noexcept nogil:
    cdef:
        double total_w = 0 # total weighting useful for interpolation
        double delta = 0 # store delta function (stopping function) result
        double new_direction[3] # new propagation direction
        double w[8]
        double qa_tmp[PEAK_NO]
        double ind_tmp[PEAK_NO]
        cnp.npy_intp index[24]
        cnp.npy_intp xyz[4]
        cnp.npy_intp i, j, m
        double normd
        # number of allowed peaks e.g. for fa is 1 for gqi.qa is 5
        cnp.npy_intp peaks = qa_shape[3]

    # Calculate qa & ind of each of the 8 neighboring voxels.
    # To do that we use trilinear interpolation and return the weights and the
    # indices for the weights i.e. xyz in qa[x,y,z]
    _trilinear_interpolation_iso(point, <double *> w, <cnp.npy_intp *> index)
    # check if you are outside of the volume
    for i from 0 <= i < 3:
        new_direction[i] = 0
        if index[7 * 3 + i] >= qa_shape[i] or index[i] < 0:
            return 0
    # for every weight sum the total weighting
    for m from 0 <= m < 8:
        for i from 0 <= i < 3:
            xyz[i]=index[m * 3 + i]
        # fill qa_tmp and ind_tmp
        for j from 0 <= j < peaks:
            xyz[3] = j
            off = offset(<cnp.npy_intp*> xyz, strides, 4, 8)
            qa_tmp[j] = qa[off]
            ind_tmp[j] = ind[off]
        # return the nearest direction by searching in all peaks
        delta=_nearest_direction(dx,
                                 qa_tmp,
                                 ind_tmp,
                                 peaks,
                                 odf_vertices,
                                 qa_thr,
                                 ang_thr,
                                 direction)
        # if delta is 0 then that means that there was no good direction
        # (obeying the thresholds) from that neighboring voxel, so this voxel
        # is not adding to the total weight
        if delta == 0:
            continue
        # add in total
        total_w += w[m]
        for i from 0 <= i < 3:
            new_direction[i] += w[m] * direction[i]
    # if less than half the volume is time to stop propagating
    if total_w < total_weight: # termination
        return 0
    # all good return normalized weighted next direction
    normd = new_direction[0]**2 + new_direction[1]**2 + new_direction[2]**2
    normd = 1 / sqrt(normd)
    for i from 0 <= i < 3:
        direction[i] = new_direction[i] * normd
    return 1


cdef cnp.npy_intp _initial_direction(double* seed,double *qa,
                                     double* ind, double* odf_vertices,
                                     double qa_thr,
                                     cnp.npy_intp* strides,
                                     cnp.npy_intp ref,
                                     double* direction) noexcept nogil:
    """ First direction that we get from a seeding point
    """
    cdef:
        cnp.npy_intp point[4]
        cnp.npy_intp off
        cnp.npy_intp i
        double qa_tmp,ind_tmp
    # Very tricky/cool addition/flooring that helps create a valid neighborhood
    # (grid) for the trilinear interpolation to run smoothly.
    # Find the index for qa
    for i from 0 <= i < 3:
        point[i] = <cnp.npy_intp>floor(seed[i] + .5)
    point[3] = ref
    # Find the offset in memory to access the qa value
    off = offset(<cnp.npy_intp*>point,strides, 4, 8)
    qa_tmp = qa[off]
    # Check for scalar threshold
    if qa_tmp < qa_thr:
        return 0
    # Find the correct direction from the indices
    ind_tmp = ind[off] # similar to ind[point] in numpy syntax
    # Return initial direction through odf_vertices by ind
    for i from 0 <= i < 3:
        direction[i] = odf_vertices[3 * <cnp.npy_intp>ind_tmp + i]
    return 1


def eudx_both_directions(cnp.ndarray[double, ndim=1] seed,
                         cnp.npy_intp ref,
                         cnp.ndarray[double, ndim=4] qa,
                         cnp.ndarray[double, ndim=4] ind,
                         cnp.ndarray[double, ndim=2] odf_vertices,
                         double qa_thr,
                         double ang_thr,
                         double step_sz,
                         double total_weight,
                         cnp.npy_intp max_points):
    """
    Parameters
    ----------
    seed : array, float64 shape (3,)
        Point where the tracking starts.
    ref : cnp.npy_intp int
        Index of peak to follow first.
    qa : array, float64 shape (X, Y, Z, Np)
        Anisotropy matrix, where ``Np`` is the number of maximum allowed peaks.
    ind : array, float64 shape(x, y, z, Np)
        Index of the track orientation.
    odf_vertices : double array shape (N, 3)
        Sampling directions on the sphere.
    qa_thr : float
        Threshold for QA, we want everything higher than this threshold.
    ang_thr : float
        Angle threshold, we only select fiber orientation within this range.
    step_sz : double
    total_weight : double
    max_points : cnp.npy_intp

    Returns
    -------
    track : array, shape (N,3)
    """
    cdef:
        double *ps = <double *> cnp.PyArray_DATA(seed)
        double *pqa = <double*> cnp.PyArray_DATA(qa)
        double *pin = <double*> cnp.PyArray_DATA(ind)
        double *pverts = <double*> cnp.PyArray_DATA(odf_vertices)
        cnp.npy_intp *pstr = <cnp.npy_intp *> cnp.PyArray_STRIDES(qa)
        cnp.npy_intp *qa_shape = <cnp.npy_intp *> cnp.PyArray_DIMS(qa)
        cnp.npy_intp *pvstr = <cnp.npy_intp *> cnp.PyArray_STRIDES(odf_vertices)
        cnp.npy_intp d, i, j, cnt
        double direction[3]
        double dx[3]
        double idirection[3]
        double ps2[3]
        double tmp, ftmp
    if not cnp.PyArray_CHKFLAGS(seed, cnp.NPY_ARRAY_C_CONTIGUOUS):
        raise ValueError("seed is not C contiguous")
    if not cnp.PyArray_CHKFLAGS(qa, cnp.NPY_ARRAY_C_CONTIGUOUS):
        raise ValueError("qa is not C contiguous")
    if not cnp.PyArray_CHKFLAGS(ind, cnp.NPY_ARRAY_C_CONTIGUOUS):
        raise ValueError("ind is not C contiguous")
    if not cnp.PyArray_CHKFLAGS(odf_vertices, cnp.NPY_ARRAY_C_CONTIGUOUS):
        raise ValueError("odf_vertices is not C contiguous")

    cnt = 0
    d = _initial_direction(ps, pqa, pin, pverts, qa_thr, pstr, ref, idirection)
    if d == 0:
        return None
    for i from 0 <= i < 3:
        # store the initial direction
        dx[i] = idirection[i]
        # ps2 is for downwards and ps for upwards propagation
        ps2[i] = ps[i]
    point = seed.copy()
    track = [point.copy()]
    # track towards one direction
    while d:
        d = _propagation_direction(ps, dx, pqa, pin, pverts, qa_thr, ang_thr,
                                   qa_shape, pstr, direction, total_weight)
        if d == 0:
            break
        if cnt > max_points:
            break
        # update the track
        for i from 0 <= i < 3:
            dx[i] = direction[i]
            # check for boundaries
            tmp = ps[i] + step_sz * dx[i]
            if tmp > qa_shape[i] - 1 or tmp < 0.:
                 d = 0
                 break
            # propagate
            ps[i] = tmp
            point[i] = ps[i]

        if d == 1:
            track.append(point.copy())
            cnt += 1
    d = 1
    for i from 0 <= i < 3:
        dx[i] = -idirection[i]

    cnt = 0
    # track towards the opposite direction
    while d:
        d = _propagation_direction(ps2, dx, pqa, pin, pverts, qa_thr, ang_thr,
                                   qa_shape, pstr, direction, total_weight)
        if d == 0:
            break
        if cnt > max_points:
            break
        # update the track
        for i from 0 <= i < 3:
            dx[i] = direction[i]
            # check for boundaries
            tmp=ps2[i] + step_sz*dx[i]
            if tmp > qa_shape[i] - 1 or tmp < 0.:
                 d = 0
                 break
            # propagate
            ps2[i] = tmp
            point[i] = ps2[i] # to be changed
        # add track point
        if d == 1:
            track.insert(0, point.copy())
            cnt += 1
    # prepare to return final track for the current seed
    tmp_track = np.array(track, dtype=np.float32)

    # Sometimes one of the ends takes small negative values; needs to be
    # investigated further

    # Return track for the current seed point and ref
    return tmp_track


cdef int initialize_ptt(TrackerParameters params,
                        double* stream_data,
                        PmfGen pmf_gen,
                        double* seed_point,
                        double* seed_direction,
                        RNGState* rng) noexcept nogil:
        """Sample an initial curve by rejection sampling.

        Parameters
        ----------
        params : TrackerParameters
            PTT tracking parameters.
        stream_data : double*
            Streamline data persitant across tracking steps.
        pmf_gen : PmfGen
            Orientation data.
        seed_point : double[3]
            Initial point
        seed_direction : double[3]
            Initial direction
        rng : RNGState*
            Random number generator state. (Threadsafe)

        Returns
        -------
        status : int
            Returns 0 if the initialization was successful, or
            1 otherwise.
        """
        cdef double data_support = 0
        cdef double max_posterior = 0
        cdef int tries

        # position
        stream_data[19] = seed_point[0]
        stream_data[20] = seed_point[1]
        stream_data[21] = seed_point[2]

        for tries in range(params.ptt.rejection_sampling_nbr_sample):
            initialize_ptt_candidate(params, stream_data, pmf_gen, seed_direction, rng)
            data_support = calculate_ptt_data_support(params, stream_data, pmf_gen)
            if data_support > max_posterior:
                max_posterior = data_support

        # Compensation for underestimation of max posterior estimate
        max_posterior = pow(2.0 * max_posterior, params.ptt.data_support_exponent)

        # Initialization is successful if a suitable candidate can be sampled
        # within the trial limit
        for tries in range(params.ptt.rejection_sampling_max_try):
            initialize_ptt_candidate(params, stream_data, pmf_gen, seed_direction, rng)
            if (random_float(rng) * max_posterior <= calculate_ptt_data_support(params, stream_data, pmf_gen)):
                stream_data[22] = stream_data[23] # last_val = last_val_cand
                return 0
        return 1


cdef void initialize_ptt_candidate(TrackerParameters params,
                                   double* stream_data,
                                   PmfGen pmf_gen,
                                   double* init_dir,
                                   RNGState* rng) noexcept nogil:
    """
    Initialize the parallel transport frame.

    After initial position is set, a parallel transport frame is set using
    the initial direction (a walking frame, i.e., 3 orthonormal vectors,
    plus 2 scalars, i.e. k1 and k2).

    A point and parallel transport frame parametrizes a curve that is named
    the "probe". Using probe parameters (probe_length, probe_radius,
    probe_quality, probe_count), a short fiber bundle segment is modelled.

    Parameters
    ----------
    params : TrackerParameters
        PTT tracking parameters.
    stream_data : double*
        Streamline data persitant across tracking steps.
    pmf_gen : PmfGen
        Orientation data.
    init_dir : double[3]
        Initial tracking direction (tangent)
    rng : RNGState*
        Random number generator state. (Threadsafe)

    Returns
    -------

    """
    cdef double[3] position
    cdef int count
    cdef int i
    cdef double* pmf

    # Initialize Frame
    stream_data[1] = init_dir[0]
    stream_data[2] = init_dir[1]
    stream_data[3] = init_dir[2]
    random_perpendicular_vector(&stream_data[7],
                                &stream_data[1],
                                rng)  # frame2, frame0
    cross(&stream_data[4],
          &stream_data[7],
          &stream_data[1])  # frame1, frame2, frame0
    stream_data[24], stream_data[25] = \
        random_point_within_circle(params.max_curvature, rng)

    stream_data[22] = 0  # last_val

    if params.ptt.probe_count == 1:
        stream_data[22] = pmf_gen.get_pmf_value_c(&stream_data[19],
                                                  &stream_data[1])  # position, frame[0]
    else:
        for count in range(params.ptt.probe_count):
            for i in range(3):
                position[i] = (stream_data[19 + i]
                              + stream_data[4 + i]
                              * params.ptt.probe_radius
                              * cos(count * params.ptt.angular_separation)
                              * params.inv_voxel_size[i]
                              + stream_data[7 + i]
                              * params.ptt.probe_radius
                              * sin(count * params.ptt.angular_separation)
                              * params.inv_voxel_size[i])

            stream_data[22] += pmf_gen.get_pmf_value_c(&stream_data[19],
                                                       &stream_data[1])


cdef void prepare_ptt_propagator(TrackerParameters params,
                                 double* stream_data,
                                 double arclength) noexcept nogil:
    """Prepare the propagator.

    The propagator used for transporting the moving frame forward.

    Parameters
    ----------
    params : TrackerParameters
        PTT tracking parameters.
    stream_data : double*
        Streamline data persitant across tracking steps.
    arclength : double
        Arclenth, which is equivalent to step size along the arc

    """
    cdef double tmp_arclength
    stream_data[10] = arclength  # propagator[0]

    if (fabs(stream_data[24]) < params.ptt.k_small
        and fabs(stream_data[25]) < params.ptt.k_small):
        stream_data[11] = 0
        stream_data[12] = 0
        stream_data[13] = 1
        stream_data[14] = 0
        stream_data[15] = 0
        stream_data[16] = 0
        stream_data[17] = 0
        stream_data[18] = 1
    else:
        if fabs(stream_data[24]) < params.ptt.k_small:  # k1
            stream_data[24] = params.ptt.k_small
        if fabs(stream_data[25]) < params.ptt.k_small:  # k2
            stream_data[25] = params.ptt.k_small

        tmp_arclength  = arclength * arclength / 2.0

        # stream_data[10:18] -> propagator
        stream_data[11] = stream_data[24] * tmp_arclength
        stream_data[12] = stream_data[25] * tmp_arclength
        stream_data[13] = (1 - stream_data[25]
                          * stream_data[25] * tmp_arclength
                          - stream_data[24] * stream_data[24]
                          * tmp_arclength)
        stream_data[14] = stream_data[24] * arclength
        stream_data[15] = stream_data[25] * arclength
        stream_data[16] = -stream_data[25] * arclength
        stream_data[17] = (-stream_data[24] * stream_data[25]
                          * tmp_arclength)
        stream_data[18] = (1 - stream_data[25] * stream_data[25]
                          * tmp_arclength)


cdef double calculate_ptt_data_support(TrackerParameters params,
                                       double* stream_data,
                                       PmfGen pmf_gen) noexcept nogil:
    """Calculates data support for the candidate probe.

    Parameters
    ----------
    params : TrackerParameters
        PTT tracking parameters.
    stream_data : double*
        Streamline data persitant across tracking steps.
    pmf_gen : PmfGen
        Orientation data.
    """

    cdef double fod_amp
    cdef double[3] position
    cdef double[3][3] frame
    cdef double[3] tangent
    cdef double[3] normal
    cdef double[3] binormal
    cdef double[3] new_position
    cdef double likelihood
    cdef int c, i, j, q
    cdef double* pmf

    prepare_ptt_propagator(params, stream_data, params.ptt.probe_step_size)

    for i in range(3):
        position[i] = stream_data[19+i]
        for j in range(3):
            frame[i][j] = stream_data[1 + i * 3 + j]

    likelihood = stream_data[22]
    for q in range(1, params.ptt.probe_quality):
        for i in range(3):
            # stream_data[10:18] : propagator
            position[i] = \
                (stream_data[10] * frame[0][i] * params.voxel_size[i]
                + stream_data[11] * frame[1][i] * params.voxel_size[i]
                + stream_data[12] * frame[2][i] * params.voxel_size[i]
                + position[i])
            tangent[i] = (stream_data[13] * frame[0][i]
                         + stream_data[14] * frame[1][i]
                         + stream_data[15] * frame[2][i])
        normalize(&tangent[0])

        if q < (params.ptt.probe_quality - 1):
            for i in range(3):
                binormal[i] = (stream_data[16] * frame[0][i]
                              + stream_data[17] * frame[1][i]
                              + stream_data[18] * frame[2][i])
            cross(&normal[0], &binormal[0], &tangent[0])

            copy_point(&tangent[0], &frame[0][0])
            copy_point(&normal[0], &frame[1][0])
            copy_point(&binormal[0], &frame[2][0])
        if params.ptt.probe_count == 1:
            fod_amp = pmf_gen.get_pmf_value_c(position, tangent)
            fod_amp = fod_amp if fod_amp > params.sh.pmf_threshold else 0
            stream_data[23] = fod_amp  # last_val_cand
            likelihood += stream_data[23]  # last_val_cand
        else:
            stream_data[23] = 0  # last_val_cand
            if q == params.ptt.probe_quality - 1:
                for i in range(3):
                    binormal[i] = (stream_data[16] * frame[0][i]
                                  + stream_data[17] * frame[1][i]
                                  + stream_data[18] * frame[2][i])
                cross(&normal[0], &binormal[0], &tangent[0])

            for c in range(params.ptt.probe_count):
                for i in range(3):
                    new_position[i] = (position[i]
                                      + normal[i] * params.ptt.probe_radius
                                      * cos(c * params.ptt.angular_separation)
                                      * params.inv_voxel_size[i]
                                      + binormal[i] * params.ptt.probe_radius
                                      * sin(c * params.ptt.angular_separation)
                                      * params.inv_voxel_size[i])
                fod_amp = pmf_gen.get_pmf_value_c(new_position, tangent)
                fod_amp = fod_amp if fod_amp > params.sh.pmf_threshold else 0
                stream_data[23] += fod_amp  # last_val_cand

            likelihood += stream_data[23]  # last_val_cand

    likelihood *= params.ptt.probe_normalizer
    if params.ptt.data_support_exponent != 1:
        likelihood = pow(likelihood, params.ptt.data_support_exponent)

    return likelihood


cdef TrackerStatus deterministic_propagator(double* point,
                                            double* direction,
                                            TrackerParameters params,
                                            double* stream_data,
                                            PmfGen pmf_gen,
                                            RNGState* rng) noexcept nogil:
    """
    Propagate the position by step_size amount.

    The propagation use the direction of a sphere with the highest probability
    mass function (pmf).

    Parameters
    ----------
    point : double[3]
        Current tracking position.
    direction : double[3]
        Previous tracking direction.
    params : TrackerParameters
        Deterministic Tractography parameters.
    stream_data : double*
        Streamline data persitant across tracking steps.
    pmf_gen : PmfGen
        Orientation data.
    rng : RNGState*
        Random number generator state. (thread safe)

    Returns
    -------
    status : TrackerStatus
        Returns SUCCESS if the propagation was successful, or
        FAIL otherwise.
    """
    cdef:
        cnp.npy_intp i, max_idx
        double max_value=0
        double* newdir
        double* pmf
        double cos_sim
        cnp.npy_intp len_pmf=pmf_gen.pmf.shape[0]

    if norm(direction) == 0:
        return FAIL
    normalize(direction)

    pmf = <double*> malloc(len_pmf * sizeof(double))
    prepare_pmf(pmf, point, pmf_gen, params.sh.pmf_threshold, len_pmf)

    for i in range(len_pmf):
        cos_sim = pmf_gen.vertices[i][0] * direction[0] \
                + pmf_gen.vertices[i][1] * direction[1] \
                + pmf_gen.vertices[i][2] * direction[2]
        if cos_sim < 0:
            cos_sim = cos_sim * -1
        if cos_sim > params.cos_similarity and pmf[i] > max_value:
            max_idx = i
            max_value = pmf[i]

    if max_value <= 0:
        free(pmf)
        return FAIL

    newdir = &pmf_gen.vertices[max_idx][0]
    # Update direction
    if (direction[0] * newdir[0]
        + direction[1] * newdir[1]
        + direction[2] * newdir[2] > 0):
        copy_point(newdir, direction)
    else:
        copy_point(newdir, direction)
        direction[0] = direction[0] * -1
        direction[1] = direction[1] * -1
        direction[2] = direction[2] * -1
    free(pmf)
    return SUCCESS


cdef TrackerStatus probabilistic_propagator(double* point,
                                            double* direction,
                                            TrackerParameters params,
                                            double* stream_data,
                                            PmfGen pmf_gen,
                                            RNGState* rng) noexcept nogil:
    """
    Propagates the position by step_size amount. The propagation use randomly samples
    direction of a sphere based on probability mass function (pmf).

    Parameters
    ----------
    point : double[3]
        Current tracking position.
    direction : double[3]
        Previous tracking direction.
    params : TrackerParameters
        Parallel Transport Tractography (PTT) parameters.
    stream_data : double*
        Streamline data persitant across tracking steps.
    pmf_gen : PmfGen
        Orientation data.
    rng : RNGState*
        Random number generator state. (thread safe)

    Returns
    -------
    status : TrackerStatus
        Returns SUCCESS if the propagation was successful, or
        FAIL otherwise.
    """

    cdef:
        cnp.npy_intp i, idx
        double* newdir
        double* pmf
        double last_cdf, cos_sim
        cnp.npy_intp len_pmf=pmf_gen.pmf.shape[0]


    if norm(direction) == 0:
        return FAIL
    normalize(direction)

    pmf = <double*> malloc(len_pmf * sizeof(double))
    prepare_pmf(pmf, point, pmf_gen, params.sh.pmf_threshold, len_pmf)

    for i in range(len_pmf):
        cos_sim = pmf_gen.vertices[i][0] * direction[0] \
                + pmf_gen.vertices[i][1] * direction[1] \
                + pmf_gen.vertices[i][2] * direction[2]
        if cos_sim < 0:
            cos_sim = cos_sim * -1
        if cos_sim < params.cos_similarity:
            pmf[i] = 0

    cumsum(pmf, pmf, len_pmf)
    last_cdf = pmf[len_pmf - 1]
    if last_cdf == 0:
        free(pmf)
        return FAIL

    idx = where_to_insert(pmf, random_float(rng) * last_cdf, len_pmf)
    newdir = &pmf_gen.vertices[idx][0]
    # Update direction
    if (direction[0] * newdir[0]
        + direction[1] * newdir[1]
        + direction[2] * newdir[2] > 0):
        copy_point(newdir, direction)
    else:
        copy_point(newdir, direction)
        direction[0] = direction[0] * -1
        direction[1] = direction[1] * -1
        direction[2] = direction[2] * -1
    free(pmf)
    return SUCCESS


cdef TrackerStatus parallel_transport_propagator(double* point,
                                              double* direction,
                                              TrackerParameters params,
                                              double* stream_data,
                                              PmfGen pmf_gen,
                                              RNGState* rng) noexcept nogil:
    """
    Propagates the position by step_size amount. The propagation is using
    the parameters of the last candidate curve. Then, randomly generate
    curve parametrization from the current position. The walking frame
    is the same, only the k1 and k2 parameters are randomly picked.
    Rejection sampling is used to pick the next curve using the data
    support (likelihood).

    stream_data:
        0    : initialized
        1-10 : frame1,2,3
        10-19: propagator
        19-22: position
        22   : last_val
        23   : last_val_cand
        24   : k1
        25   : k2

    Parameters
    ----------
    point : double[3]
        Current tracking position.
    direction : double[3]
        Previous tracking direction.
    params : TrackerParameters
        Parallel Transport Tractography (PTT) parameters.
    stream_data : double*
        Streamline data persitant across tracking steps.
    pmf_gen : PmfGen
        Orientation data.
    rng : RNGState*
        Random number generator state. (thread safe)

    Returns
    -------
    status : TrackerStatus
        Returns SUCCESS if the propagation was successful, or
        FAIL otherwise.
    """

    cdef double max_posterior = 0
    cdef double data_support = 0
    cdef double[3] tangent
    cdef int tries
    cdef int i

    if stream_data[0] == 0:
        initialize_ptt(params, stream_data, pmf_gen, point, direction, rng)
        stream_data[0] = 1  # initialized

    prepare_ptt_propagator(params, stream_data, params.step_size)

    for i in range(3):
        #  position
        stream_data[19 + i] = (stream_data[10] * stream_data[1 + i]
                               * params.inv_voxel_size[i]
                               + stream_data[11] * stream_data[4 + i]
                               * params.inv_voxel_size[i]
                               + stream_data[12] * stream_data[7 + i]
                               * params.inv_voxel_size[i]
                               + stream_data[19 + i])
        tangent[i] = (stream_data[13] * stream_data[1 + i]
                      + stream_data[14] * stream_data[4 + i]
                      + stream_data[15] * stream_data[7 + i])
        stream_data[7 + i] = \
            (stream_data[16] * stream_data[1 + i]
            + stream_data[17] * stream_data[4 + i]
            + stream_data[18] * stream_data[7 + i])
    normalize(&tangent[0])
    cross(&stream_data[4], &stream_data[7], &tangent[0])  # frame1, frame2
    normalize(&stream_data[4])  # frame1
    cross(&stream_data[7], &tangent[0], &stream_data[4])  # frame2, tangent, frame1
    stream_data[1] = tangent[0]
    stream_data[2] = tangent[1]
    stream_data[3] = tangent[2]

    for tries in range(params.ptt.rejection_sampling_nbr_sample):
        # k1, k2
        stream_data[24], stream_data[25] = \
            random_point_within_circle(params.max_curvature, rng)
        data_support = calculate_ptt_data_support(params, stream_data, pmf_gen)
        if data_support > max_posterior:
            max_posterior = data_support

    # Compensation for underestimation of max posterior estimate
    max_posterior = pow(2.0 * max_posterior, params.ptt.data_support_exponent)


    for tries in range(params.ptt.rejection_sampling_max_try):
        # k1, k2
        stream_data[24], stream_data[25] = \
            random_point_within_circle(params.max_curvature, rng)
        if random_float(rng) * max_posterior < calculate_ptt_data_support(params, stream_data, pmf_gen):
            stream_data[22] = stream_data[23] # last_val = last_val_cand
            # Propagation is successful if a suitable candidate can be sampled
            # within the trial limit
            # update the point and return
            copy_point(&stream_data[19], point)
            return SUCCESS

    return FAIL



"""
======================================================
Denoise images using the Marcenko-Pastur PCA algorithm
======================================================

The PCA-based denoising algorithm exploits the redundancy across the
diffusion-weighted images :footcite:p:`Manjon2013`, :footcite:p:`Veraart2016b`.
This algorithm has been shown to provide an optimal compromise between noise
suppression and loss of anatomical information for different techniques such as
DTI :footcite:p:`Manjon2013`, spherical deconvolution :footcite:p:`Veraart2016b`
and DKI :footcite:p:`NetoHenriques2018`.

The basic idea behind the PCA-based denoising algorithms is to remove the
components of the data that are classified as noise. The Principal Components
classification can be performed based on prior noise variance estimates
:footcite:p:`Manjon2013`
(see :ref:`denoise_localpca<sphx_glr_examples_built_preprocessing_denoise_localpca.py>`)
or automatically based on the Marchenko-Pastur distribution
:footcite:p:`Veraart2016b`. In addition to noise suppression, the PCA algorithm
can be used to get the standard deviation of the noise
:footcite:p:`Veraart2016b`.

In the following example, we show how to denoise diffusion MRI images and
estimate the noise standard deviation using the PCA algorithm based
on the Marcenko-Pastur distribution :footcite:p:`Veraart2016b`

Let's load the necessary modules
"""  # noqa: E501

# load general modules
from time import time

import matplotlib.pyplot as plt
import numpy as np

# load other dipy's functions that will be used for auxiliary analysis
from dipy.core.gradients import gradient_table

# load functions to fetch data for this example
from dipy.data import get_fnames

# load main pca function using Marcenko-Pastur distribution
from dipy.denoise.localpca import mppca
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti, save_nifti
import dipy.reconst.dki as dki
from dipy.segment.mask import median_otsu

###############################################################################
# For this example, we use fetch to download a multi-shell dataset which was
# kindly provided by Hansen and Jespersen (more details about the data are
# provided in their paper :footcite:p:`Hansen2016a`). The total size of the
# downloaded data is 192 MBytes, however you only need to fetch it once.

dwi_fname, dwi_bval_fname, dwi_bvec_fname, _ = get_fnames(name="cfin_multib")
data, affine = load_nifti(dwi_fname)
bvals, bvecs = read_bvals_bvecs(dwi_bval_fname, dwi_bvec_fname)
gtab = gradient_table(bvals, bvecs=bvecs)

###############################################################################
# For the sake of simplicity, we only select two non-zero b-values for this
# example.

bvals = gtab.bvals

bvecs = gtab.bvecs

sel_b = np.logical_or(np.logical_or(bvals == 0, bvals == 1000), bvals == 2000)

data = data[..., sel_b]

gtab = gradient_table(bvals[sel_b], bvecs=bvecs[sel_b])

print(data.shape)

###############################################################################
# As one can see from its shape, the selected data contains a total of 67
# volumes of images corresponding to all the diffusion gradient directions
# of the selected b-values.
#
# The PCA denoising using the Marchenko-Pastur distribution can be performed by
# calling the function ``mppca``:

t = time()

denoised_arr = mppca(data, patch_radius=2)

print("Time taken for local MP-PCA ", -t + time())

###############################################################################
# Internally, the ``mppca`` algorithm denoises the diffusion-weighted data
# using a 3D sliding window which is defined by the variable patch_radius.
# In total, this window should comprise a larger number of voxels than the
# number of diffusion-weighted volumes. Since our data has a total of 67
# volumes, the patch_radius is set to 2 which corresponds to a 5x5x5 sliding
# window comprising a total of 125 voxels.

# To assess the performance of the Marchenko-Pastur PCA denoising algorithm,
# an axial slice of the original data, denoised data, and residuals are
# plotted below:

sli = data.shape[2] // 2
gra = data.shape[3] - 1
orig = data[:, :, sli, gra]
den = denoised_arr[:, :, sli, gra]
rms_diff = np.sqrt((orig - den) ** 2)

fig1, ax = plt.subplots(1, 3, figsize=(12, 6), subplot_kw={"xticks": [], "yticks": []})

fig1.subplots_adjust(hspace=0.3, wspace=0.05)

ax.flat[0].imshow(orig.T, cmap="gray", interpolation="none", origin="lower")
ax.flat[0].set_title("Original")
ax.flat[1].imshow(den.T, cmap="gray", interpolation="none", origin="lower")
ax.flat[1].set_title("Denoised Output")
ax.flat[2].imshow(rms_diff.T, cmap="gray", interpolation="none", origin="lower")
ax.flat[2].set_title("Residuals")

fig1.savefig("denoised_mppca.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# The noise suppression can be visually appreciated by comparing the original
# data slice (left panel) to its denoised version (middle panel). The
# difference between original and denoised data showing only random noise
# indicates that the data's structural information is preserved by the PCA
# denoising algorithm (right panel).
#
#
# Below we show how the denoised data can be saved.

save_nifti("denoised_mppca.nii.gz", denoised_arr, affine)

###############################################################################
# Additionally, we show how the PCA denoising algorithm affects different
# diffusion measurements. For this, we run the diffusion kurtosis model
# below on both original and denoised versions of the data:

dkimodel = dki.DiffusionKurtosisModel(gtab)

maskdata, mask = median_otsu(
    data, vol_idx=[0, 1], median_radius=4, numpass=2, autocrop=False, dilate=1
)

dki_orig = dkimodel.fit(data, mask=mask)
dki_den = dkimodel.fit(denoised_arr, mask=mask)

###############################################################################
# We use the following code to plot the MD, FA and MK estimates from the two
# data fits:

FA_orig = dki_orig.fa
FA_den = dki_den.fa
MD_orig = dki_orig.md
MD_den = dki_den.md
MK_orig = dki_orig.mk(min_kurtosis=0, max_kurtosis=3)
MK_den = dki_den.mk(min_kurtosis=0, max_kurtosis=3)


fig2, ax = plt.subplots(2, 3, figsize=(10, 6), subplot_kw={"xticks": [], "yticks": []})

fig2.subplots_adjust(hspace=0.3, wspace=0.03)

ax.flat[0].imshow(
    MD_orig[:, :, sli].T, cmap="gray", vmin=0, vmax=2.0e-3, origin="lower"
)
ax.flat[0].set_title("MD (DKI)")
ax.flat[1].imshow(FA_orig[:, :, sli].T, cmap="gray", vmin=0, vmax=0.7, origin="lower")
ax.flat[1].set_title("FA (DKI)")
ax.flat[2].imshow(MK_orig[:, :, sli].T, cmap="gray", vmin=0, vmax=1.5, origin="lower")
ax.flat[2].set_title("AD (DKI)")
ax.flat[3].imshow(MD_den[:, :, sli].T, cmap="gray", vmin=0, vmax=2.0e-3, origin="lower")
ax.flat[3].set_title("MD (DKI)")
ax.flat[4].imshow(FA_den[:, :, sli].T, cmap="gray", vmin=0, vmax=0.7, origin="lower")
ax.flat[4].set_title("FA (DKI)")
ax.flat[5].imshow(MK_den[:, :, sli].T, cmap="gray", vmin=0, vmax=1.5, origin="lower")
ax.flat[5].set_title("AD (DKI)")
plt.show()
fig2.savefig("denoised_dki.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# In the above figure, the DKI maps obtained from the original data are shown
# in the upper panels, while the DKI maps from the denoised data are shown in
# the lower panels. Substantial improvements in measurement robustness can be
# visually appreciated, particularly for the FA and MK estimates.
#
#
# Noise standard deviation estimation using the Marchenko-Pastur PCA algorithm
# ============================================================================
#
# As mentioned above, the Marcenko-Pastur PCA algorithm can also be used to
# estimate the image's noise standard deviation (std). The noise std
# automatically computed from the ``mppca`` function can be returned by
# setting the optional input parameter ``return_sigma`` to True.

denoised_arr, sigma = mppca(data, patch_radius=2, return_sigma=True)

###############################################################################
# Let's plot the noise standard deviation estimate:

fig3 = plt.figure("PCA Noise standard deviation estimation")
plt.imshow(sigma[..., sli].T, cmap="gray", origin="lower")
plt.axis("off")
plt.show()
fig3.savefig("pca_sigma.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# The above figure shows that the Marchenko-Pastur PCA algorithm computes a 3D
# spatial varying noise level map. To obtain the mean noise std across all
# voxels, you can use the following lines of code:

mean_sigma = np.mean(sigma[mask])

print(mean_sigma)

###############################################################################
# Below we use this mean noise level estimate to compute the nominal SNR of
# the data (i.e. SNR at b-value=0):

b0 = denoised_arr[..., 0]

mean_signal = np.mean(b0[mask])

snr = mean_signal / mean_sigma

print(snr)

###############################################################################
# References
# ----------
#
# .. footbibliography::
#



import numpy as np
from numpy.testing import assert_almost_equal, assert_raises

from dipy.data import get_fnames
from dipy.io.streamline import load_tractogram
from dipy.segment.fss import (
    FastStreamlineSearch,
    nearest_from_matrix_col,
    nearest_from_matrix_row,
)
from dipy.segment.metric import mean_euclidean_distance
from dipy.testing import (
    assert_arrays_equal,
    assert_greater,
    assert_greater_equal,
    assert_true,
)


def setup_module():
    global f1, f2
    fname = get_fnames(name="fornix")
    fornix = load_tractogram(fname, "same", bbox_valid_check=False)

    # Should work with both StatefulTractogram and streamlines (list of array)
    f1 = fornix[:200]
    f2 = fornix.streamlines[200:]


def test_fss_radius_search():
    r = 4.0
    nb_pts = 24
    # For each "f1" streamlines search all in radius of "f2"
    fss_f1 = FastStreamlineSearch(
        f1,
        max_radius=r,
        nb_mpts=2,
        bin_size=20.0,
        resampling=nb_pts,
        bidirectional=True,
    )
    rs_f2_in_f1 = fss_f1.radius_search(f2, radius=r, use_negative=True)

    # For each "f2" streamlines search all in radius of "f1"
    fss_f2 = FastStreamlineSearch(
        f2,
        max_radius=r,
        nb_mpts=8,
        bin_size=10.0,
        resampling=nb_pts,
        bidirectional=True,
    )
    rs_f1_in_f2 = fss_f2.radius_search(f1, radius=r, use_negative=False)

    # Verify there is results
    assert_greater(rs_f2_in_f1.nnz, 0)
    assert_greater(rs_f1_in_f2.nnz, 0)

    # Verify The number of results are the same
    assert_true(rs_f2_in_f1.nnz == rs_f1_in_f2.nnz)

    # Verify that both search results are equivalent (transposed)
    assert_arrays_equal(np.sort(rs_f2_in_f1.row), np.sort(rs_f1_in_f2.col))
    assert_arrays_equal(np.sort(rs_f2_in_f1.col), np.sort(rs_f1_in_f2.row))

    # Verify if resulting distances are the same
    assert_almost_equal(np.sort(np.abs(rs_f2_in_f1.data)), np.sort(rs_f1_in_f2.data))

    # Verify that minimum are the same from f1 to f2
    r1_a, r1_b, r1_d = nearest_from_matrix_row(rs_f2_in_f1)
    r2_a, r2_b, r2_d = nearest_from_matrix_col(rs_f1_in_f2)
    assert_arrays_equal(r1_a, r2_a)
    assert_arrays_equal(r1_b, r2_b)
    assert_arrays_equal(r1_d, r2_d)

    # Verify that minimum are the same from f2 to f1
    r3_a, r3_b, r3_d = nearest_from_matrix_col(rs_f2_in_f1)
    r4_a, r4_b, r4_d = nearest_from_matrix_row(rs_f1_in_f2)
    assert_arrays_equal(r3_a, r4_a)
    assert_arrays_equal(r3_b, r4_b)
    assert_almost_equal(r3_d, r4_d)

    # Test with unidirectional search (bidirectional=False)
    fss_sd = FastStreamlineSearch(
        f1,
        max_radius=r,
        nb_mpts=6,
        bin_size=80.0,
        resampling=nb_pts,
        bidirectional=False,
    )
    rs_f1_sd = fss_sd.radius_search(f2, radius=r, use_negative=True)

    # Single direction should be a subset of bidirectional
    assert_greater_equal(rs_f1_in_f2.nnz, rs_f1_sd.nnz)
    assert_true(np.all(np.isin(rs_f1_sd.row, rs_f2_in_f1.row)))
    assert_true(np.all(np.isin(rs_f1_sd.col, rs_f2_in_f1.col)))


def test_fss_varying_radius():
    # For each "f1" streamlines search all in radius of "f2"
    fss = FastStreamlineSearch(
        f1, max_radius=10.0, nb_mpts=5, bin_size=20.0, resampling=25, bidirectional=True
    )
    rs_6 = fss.radius_search(f2, radius=6.0, use_negative=True)
    rs_4 = fss.radius_search(f2, radius=4.0, use_negative=True)
    rs_2 = fss.radius_search(f2, radius=2.0, use_negative=True)

    # smaller radius should be a subset or equal of a bigger radius
    assert_greater_equal(rs_6.nnz, rs_4.nnz)
    assert_true(np.all(np.isin(rs_4.row, rs_6.row)))
    assert_true(np.all(np.isin(rs_4.col, rs_6.col)))

    assert_greater_equal(rs_4.nnz, rs_2.nnz)
    assert_true(np.all(np.isin(rs_2.row, rs_4.row)))
    assert_true(np.all(np.isin(rs_2.col, rs_4.col)))


def test_fss_single_point_slines():
    slines = [np.array([[1.0, 1.0, 1.0]]), np.array([[0.0, 1.0, 2.0]])]
    fss = FastStreamlineSearch(
        slines,
        max_radius=4.0,
        nb_mpts=4,
        bin_size=20.0,
        resampling=24,
        bidirectional=False,
    )
    res = fss.radius_search(slines, radius=4.0)
    # 2x2 matrix with 4 element
    assert_true(res.nnz == 4)
    mat = res.toarray()
    dist = mean_euclidean_distance(slines[0], slines[1])
    assert_almost_equal(mat[0, 0], 0.0)
    assert_almost_equal(mat[1, 1], 0.0)
    assert_almost_equal(mat[1, 0], dist)
    assert_almost_equal(mat[0, 1], dist)


def test_fss_empty_results():
    fss = FastStreamlineSearch(
        f1, max_radius=2.0, nb_mpts=2, bin_size=20.0, resampling=22, bidirectional=True
    )
    res = fss.radius_search(f2, radius=0.01, use_negative=True)
    assert_true(res.nnz == 0)


def test_fss_invalid_max_radius():
    assert_raises(ValueError, FastStreamlineSearch, f1, max_radius=0.0)
    assert_raises(ValueError, FastStreamlineSearch, f1, max_radius=-1.0)


def test_fss_invalid_radius():
    fss = FastStreamlineSearch(f1, max_radius=2.0)
    assert_raises(ValueError, fss.radius_search, f2, radius=5.0)


def test_fss_invalid_mpts():
    assert_raises(
        ValueError, FastStreamlineSearch, f1, max_radius=4.0, nb_mpts=4, resampling=23
    )
    assert_raises(
        ZeroDivisionError,
        FastStreamlineSearch,
        f1,
        max_radius=4.0,
        nb_mpts=0,
        resampling=24,
    )



from dipy.segment.cythonutils cimport Data2D, Shape
from dipy.segment.featurespeed cimport Feature


cdef class Metric:
    cdef Feature feature
    cdef int is_order_invariant

    cdef double c_dist(Metric self, Data2D features1, Data2D features2) except -1 nogil
    cdef int c_are_compatible(Metric self, Shape shape1, Shape shape2) except -1 nogil

    cpdef double dist(Metric self, features1, features2) except -1
    cpdef are_compatible(Metric self, shape1, shape2)



import numpy as np
import numpy.testing as npt

from dipy.data import get_fnames
from dipy.segment.mrf import ConstantObservationModel, IteratedConditionalModes
from dipy.segment.tissue import TissueClassifierHMRF
from dipy.sims.voxel import add_noise
from dipy.testing.decorators import set_random_number_generator


def create_image():
    # Load a coronal slice from a T1-weighted MRI
    fname = get_fnames(name="t1_coronal_slice")
    single_slice = np.load(fname)

    # Stack a few copies to form a 3D volume
    nslices = 5
    image = np.zeros(shape=single_slice.shape + (nslices,))
    image[..., :nslices] = single_slice[..., None]
    return image


# Making squares
def create_square():
    square = np.zeros((256, 256, 3), dtype=np.int16)
    square[42:213, 42:213, :] = 1
    square[71:185, 71:185, :] = 2
    square[99:157, 99:157, :] = 3

    return square


def create_square_uniform(rng):
    square_1 = np.zeros((256, 256, 3)) + 0.001
    square_1 = add_noise(square_1, 10000, 1, noise_type="gaussian", rng=rng)
    temp_1 = rng.integers(1, 21, size=(171, 171, 3))
    temp_1 = np.where(temp_1 < 20, 1, 3)
    square_1[42:213, 42:213, :] = temp_1
    temp_2 = rng.integers(1, 21, size=(114, 114, 3))
    temp_2 = np.where(temp_2 < 19, 2, 1)
    square_1[71:185, 71:185, :] = temp_2
    temp_3 = rng.integers(1, 21, size=(58, 58, 3))
    temp_3 = np.where(temp_3 < 20, 3, 1)
    square_1[99:157, 99:157, :] = temp_3

    return square_1


def create_square_gauss(rng):
    square_gauss = np.zeros((256, 256, 3)) + 0.001
    square_gauss = add_noise(square_gauss, 10000, 1, noise_type="gaussian", rng=rng)
    square_gauss[42:213, 42:213, :] = 1
    noise_1 = rng.normal(1.001, 0.0001, size=square_gauss[42:213, 42:213, :].shape)
    square_gauss[42:213, 42:213, :] = square_gauss[42:213, 42:213, :] + noise_1
    square_gauss[71:185, 71:185, :] = 2
    noise_2 = rng.normal(2.001, 0.0001, size=square_gauss[71:185, 71:185, :].shape)
    square_gauss[71:185, 71:185, :] = square_gauss[71:185, 71:185, :] + noise_2
    square_gauss[99:157, 99:157, :] = 3
    noise_3 = rng.normal(3.001, 0.0001, size=square_gauss[99:157, 99:157, :].shape)
    square_gauss[99:157, 99:157, :] = square_gauss[99:157, 99:157, :] + noise_3

    return square_gauss


def test_grayscale_image():
    nclasses = 4
    beta = np.float64(0.0)

    image = create_image()
    com = ConstantObservationModel()
    icm = IteratedConditionalModes()

    mu, sigmasq = com.initialize_param_uniform(image, nclasses)
    npt.assert_array_almost_equal(mu, np.array([0.0, 0.25, 0.5, 0.75]))
    npt.assert_array_almost_equal(sigmasq, np.array([1.0, 1.0, 1.0, 1.0]))

    neglogl = com.negloglikelihood(image, mu, sigmasq, nclasses)
    npt.assert_(neglogl[100, 100, 1, 0] != neglogl[100, 100, 1, 1])
    npt.assert_(neglogl[100, 100, 1, 1] != neglogl[100, 100, 1, 2])
    npt.assert_(neglogl[100, 100, 1, 2] != neglogl[100, 100, 1, 3])
    npt.assert_(neglogl[100, 100, 1, 1] != neglogl[100, 100, 1, 3])

    initial_segmentation = icm.initialize_maximum_likelihood(neglogl)
    npt.assert_(initial_segmentation.max() == nclasses - 1)
    npt.assert_(initial_segmentation.min() == 0)

    PLN = icm.prob_neighborhood(initial_segmentation, beta, nclasses)
    print(PLN.shape)
    npt.assert_(np.all((PLN >= 0) & (PLN <= 1.0)))

    if beta == 0.0:
        npt.assert_almost_equal(PLN[50, 50, 1, 0], 0.25, True)
        npt.assert_almost_equal(PLN[50, 50, 1, 1], 0.25, True)
        npt.assert_almost_equal(PLN[50, 50, 1, 2], 0.25, True)
        npt.assert_almost_equal(PLN[50, 50, 1, 3], 0.25, True)
        npt.assert_almost_equal(PLN[147, 129, 1, 0], 0.25, True)
        npt.assert_almost_equal(PLN[147, 129, 1, 1], 0.25, True)
        npt.assert_almost_equal(PLN[147, 129, 1, 2], 0.25, True)
        npt.assert_almost_equal(PLN[147, 129, 1, 3], 0.25, True)
        npt.assert_almost_equal(PLN[61, 152, 1, 0], 0.25, True)
        npt.assert_almost_equal(PLN[61, 152, 1, 1], 0.25, True)
        npt.assert_almost_equal(PLN[61, 152, 1, 2], 0.25, True)
        npt.assert_almost_equal(PLN[61, 152, 1, 3], 0.25, True)
        npt.assert_almost_equal(PLN[100, 100, 1, 0], 0.25, True)
        npt.assert_almost_equal(PLN[100, 100, 1, 1], 0.25, True)
        npt.assert_almost_equal(PLN[100, 100, 1, 2], 0.25, True)
        npt.assert_almost_equal(PLN[100, 100, 1, 3], 0.25, True)

    PLY = com.prob_image(image, nclasses, mu, sigmasq, PLN)
    print(PLY)
    npt.assert_(np.all((PLY >= 0) & (PLY <= 1.0)))

    mu_upd, sigmasq_upd = com.update_param(image, PLY, mu, nclasses)
    print(mu)
    print(mu_upd)
    npt.assert_(mu_upd[0] != mu[0])
    npt.assert_(mu_upd[1] != mu[1])
    npt.assert_(mu_upd[2] != mu[2])
    npt.assert_(mu_upd[3] != mu[3])
    print(sigmasq)
    print(sigmasq_upd)
    npt.assert_(sigmasq_upd[0] != sigmasq[0])
    npt.assert_(sigmasq_upd[1] != sigmasq[1])
    npt.assert_(sigmasq_upd[2] != sigmasq[2])
    npt.assert_(sigmasq_upd[3] != sigmasq[3])

    icm_segmentation, energy = icm.icm_ising(neglogl, beta, initial_segmentation)
    npt.assert_(np.abs(np.sum(icm_segmentation)) != 0)
    npt.assert_(icm_segmentation.max() == nclasses - 1)
    npt.assert_(icm_segmentation.min() == 0)


def test_grayscale_iter():
    nclasses = 4
    beta = np.float64(0.1)
    max_iter = 15
    background_noise = True

    image = create_image()
    com = ConstantObservationModel()
    icm = IteratedConditionalModes()

    mu, sigmasq = com.initialize_param_uniform(image, nclasses)
    neglogl = com.negloglikelihood(image, mu, sigmasq, nclasses)
    initial_segmentation = icm.initialize_maximum_likelihood(neglogl)
    npt.assert_(initial_segmentation.max() == nclasses - 1)
    npt.assert_(initial_segmentation.min() == 0)

    mu, sigmasq = com.seg_stats(image, initial_segmentation, nclasses)
    npt.assert_(mu[0] >= 0.0)
    npt.assert_(mu[1] >= 0.0)
    npt.assert_(mu[2] >= 0.0)
    npt.assert_(mu[3] >= 0.0)
    npt.assert_(sigmasq[0] >= 0.0)
    npt.assert_(sigmasq[1] >= 0.0)
    npt.assert_(sigmasq[2] >= 0.0)
    npt.assert_(sigmasq[3] >= 0.0)

    if background_noise:
        zero = np.zeros_like(image) + 0.001
        zero_noise = add_noise(zero, 10000, 1, noise_type="gaussian")
        image_gauss = np.where(image == 0, zero_noise, image)
    else:
        image_gauss = image

    final_segmentation = np.empty_like(image)
    seg_init = initial_segmentation
    energies = []

    for _ in range(max_iter):
        PLN = icm.prob_neighborhood(initial_segmentation, beta, nclasses)
        npt.assert_(np.all((PLN >= 0) & (PLN <= 1.0)))

        if beta == 0.0:
            npt.assert_almost_equal(PLN[50, 50, 1, 0], 0.25, True)
            npt.assert_almost_equal(PLN[50, 50, 1, 1], 0.25, True)
            npt.assert_almost_equal(PLN[50, 50, 1, 2], 0.25, True)
            npt.assert_almost_equal(PLN[50, 50, 1, 3], 0.25, True)
            npt.assert_almost_equal(PLN[147, 129, 1, 0], 0.25, True)
            npt.assert_almost_equal(PLN[147, 129, 1, 1], 0.25, True)
            npt.assert_almost_equal(PLN[147, 129, 1, 2], 0.25, True)
            npt.assert_almost_equal(PLN[147, 129, 1, 3], 0.25, True)
            npt.assert_almost_equal(PLN[61, 152, 1, 0], 0.25, True)
            npt.assert_almost_equal(PLN[61, 152, 1, 1], 0.25, True)
            npt.assert_almost_equal(PLN[61, 152, 1, 2], 0.25, True)
            npt.assert_almost_equal(PLN[61, 152, 1, 3], 0.25, True)
            npt.assert_almost_equal(PLN[100, 100, 1, 0], 0.25, True)
            npt.assert_almost_equal(PLN[100, 100, 1, 1], 0.25, True)
            npt.assert_almost_equal(PLN[100, 100, 1, 2], 0.25, True)
            npt.assert_almost_equal(PLN[100, 100, 1, 3], 0.25, True)

        PLY = com.prob_image(image_gauss, nclasses, mu, sigmasq, PLN)
        npt.assert_(np.all((PLY >= 0) & (PLY <= 1.0)))
        npt.assert_(PLY[50, 50, 1, 0] > PLY[50, 50, 1, 1])
        npt.assert_(PLY[50, 50, 1, 0] > PLY[50, 50, 1, 2])
        npt.assert_(PLY[50, 50, 1, 0] > PLY[50, 50, 1, 3])
        npt.assert_(PLY[100, 100, 1, 3] > PLY[100, 100, 1, 0])
        npt.assert_(PLY[100, 100, 1, 3] > PLY[100, 100, 1, 1])
        npt.assert_(PLY[100, 100, 1, 3] > PLY[100, 100, 1, 2])

        mu_upd, sigmasq_upd = com.update_param(image_gauss, PLY, mu, nclasses)
        npt.assert_(mu_upd[0] >= 0.0)
        npt.assert_(mu_upd[1] >= 0.0)
        npt.assert_(mu_upd[2] >= 0.0)
        npt.assert_(mu_upd[3] >= 0.0)
        npt.assert_(sigmasq_upd[0] >= 0.0)
        npt.assert_(sigmasq_upd[1] >= 0.0)
        npt.assert_(sigmasq_upd[2] >= 0.0)
        npt.assert_(sigmasq_upd[3] >= 0.0)

        negll = com.negloglikelihood(image_gauss, mu_upd, sigmasq_upd, nclasses)
        npt.assert_(negll[50, 50, 1, 0] < negll[50, 50, 1, 1])
        npt.assert_(negll[50, 50, 1, 0] < negll[50, 50, 1, 2])
        npt.assert_(negll[50, 50, 1, 0] < negll[50, 50, 1, 3])
        npt.assert_(negll[100, 100, 1, 3] < negll[100, 100, 1, 0])
        npt.assert_(negll[100, 100, 1, 3] < negll[100, 100, 1, 1])
        npt.assert_(negll[100, 100, 1, 3] < negll[100, 100, 1, 2])

        final_segmentation, energy = icm.icm_ising(negll, beta, initial_segmentation)
        print(energy[energy > -np.inf].sum())
        energies.append(energy[energy > -np.inf].sum())

        initial_segmentation = final_segmentation
        mu = mu_upd
        sigmasq = sigmasq_upd

    npt.assert_(energies[-1] < energies[0])

    difference_map = np.abs(seg_init - final_segmentation)
    npt.assert_(np.abs(np.sum(difference_map)) != 0)


@set_random_number_generator()
def test_square_iter(rng):
    nclasses = 4
    beta = np.float64(0.0)
    max_iter = 10

    com = ConstantObservationModel()
    icm = IteratedConditionalModes()

    initial_segmentation = create_square()
    square_gauss = create_square_gauss(rng)

    mu, sigmasq = com.seg_stats(square_gauss, initial_segmentation, nclasses)
    npt.assert_(mu[0] >= 0.0)
    npt.assert_(mu[1] >= 0.0)
    npt.assert_(mu[2] >= 0.0)
    npt.assert_(mu[3] >= 0.0)
    npt.assert_(sigmasq[0] >= 0.0)
    npt.assert_(sigmasq[1] >= 0.0)
    npt.assert_(sigmasq[2] >= 0.0)
    npt.assert_(sigmasq[3] >= 0.0)

    final_segmentation = np.empty_like(square_gauss)
    seg_init = initial_segmentation
    energies = []

    for i in range(max_iter):
        print("\n")
        print(f">> Iteration: {i}")
        print("\n")

        PLN = icm.prob_neighborhood(initial_segmentation, beta, nclasses)
        npt.assert_(np.all((PLN >= 0) & (PLN <= 1.0)))

        if beta == 0.0:
            npt.assert_(PLN[25, 25, 1, 0] == 0.25)
            npt.assert_(PLN[25, 25, 1, 1] == 0.25)
            npt.assert_(PLN[25, 25, 1, 2] == 0.25)
            npt.assert_(PLN[25, 25, 1, 3] == 0.25)
            npt.assert_(PLN[50, 50, 1, 0] == 0.25)
            npt.assert_(PLN[50, 50, 1, 1] == 0.25)
            npt.assert_(PLN[50, 50, 1, 2] == 0.25)
            npt.assert_(PLN[50, 50, 1, 3] == 0.25)
            npt.assert_(PLN[90, 90, 1, 0] == 0.25)
            npt.assert_(PLN[90, 90, 1, 1] == 0.25)
            npt.assert_(PLN[90, 90, 1, 2] == 0.25)
            npt.assert_(PLN[90, 90, 1, 3] == 0.25)
            npt.assert_(PLN[125, 125, 1, 0] == 0.25)
            npt.assert_(PLN[125, 125, 1, 1] == 0.25)
            npt.assert_(PLN[125, 125, 1, 2] == 0.25)
            npt.assert_(PLN[125, 125, 1, 3] == 0.25)

        PLY = com.prob_image(square_gauss, nclasses, mu, sigmasq, PLN)
        npt.assert_(np.all((PLY >= 0) & (PLY <= 1.0)))
        npt.assert_(PLY[25, 25, 1, 0] > PLY[25, 25, 1, 1])
        npt.assert_(PLY[25, 25, 1, 0] > PLY[25, 25, 1, 2])
        npt.assert_(PLY[25, 25, 1, 0] > PLY[25, 25, 1, 3])
        npt.assert_(PLY[125, 125, 1, 3] > PLY[125, 125, 1, 0])
        npt.assert_(PLY[125, 125, 1, 3] > PLY[125, 125, 1, 1])
        npt.assert_(PLY[125, 125, 1, 3] > PLY[125, 125, 1, 2])

        mu_upd, sigmasq_upd = com.update_param(square_gauss, PLY, mu, nclasses)
        npt.assert_(mu_upd[0] >= 0.0)
        npt.assert_(mu_upd[1] >= 0.0)
        npt.assert_(mu_upd[2] >= 0.0)
        npt.assert_(mu_upd[3] >= 0.0)
        npt.assert_(sigmasq_upd[0] >= 0.0)
        npt.assert_(sigmasq_upd[1] >= 0.0)
        npt.assert_(sigmasq_upd[2] >= 0.0)
        npt.assert_(sigmasq_upd[3] >= 0.0)

        negll = com.negloglikelihood(square_gauss, mu_upd, sigmasq_upd, nclasses)
        npt.assert_(negll[25, 25, 1, 0] < negll[25, 25, 1, 1])
        npt.assert_(negll[25, 25, 1, 0] < negll[25, 25, 1, 2])
        npt.assert_(negll[25, 25, 1, 0] < negll[25, 25, 1, 3])
        npt.assert_(negll[100, 100, 1, 3] < negll[125, 125, 1, 0])
        npt.assert_(negll[100, 100, 1, 3] < negll[125, 125, 1, 1])
        npt.assert_(negll[100, 100, 1, 3] < negll[125, 125, 1, 2])

        final_segmentation, energy = icm.icm_ising(negll, beta, initial_segmentation)

        energies.append(energy[energy > -np.inf].sum())

        initial_segmentation = final_segmentation
        mu = mu_upd
        sigmasq = sigmasq_upd

    difference_map = np.abs(seg_init - final_segmentation)
    npt.assert_(np.abs(np.sum(difference_map)) == 0.0)


@set_random_number_generator()
def test_icm_square(rng):
    nclasses = 4
    max_iter = 10

    com = ConstantObservationModel()
    icm = IteratedConditionalModes()

    initial_segmentation = create_square()
    square_1 = create_square_uniform(rng)

    mu, sigma = com.seg_stats(square_1, initial_segmentation, nclasses)
    sigmasq = sigma**2
    npt.assert_(mu[0] >= 0.0)
    npt.assert_(mu[1] >= 0.0)
    npt.assert_(mu[2] >= 0.0)
    npt.assert_(mu[3] >= 0.0)
    npt.assert_(sigmasq[0] >= 0.0)
    npt.assert_(sigmasq[1] >= 0.0)
    npt.assert_(sigmasq[2] >= 0.0)
    npt.assert_(sigmasq[3] >= 0.0)

    negll = com.negloglikelihood(square_1, mu, sigmasq, nclasses)

    final_segmentation_1 = np.empty_like(square_1)
    final_segmentation_2 = np.empty_like(square_1)

    beta = 0.0

    for i in range(max_iter):
        print("\n")
        print(f">> Iteration: {i}")
        print("\n")

        final_segmentation_1, energy_1 = icm.icm_ising(
            negll, beta, initial_segmentation
        )
        initial_segmentation = final_segmentation_1

    beta = 2
    initial_segmentation = create_square()

    for j in range(max_iter):
        print("\n")
        print(f">> Iteration: {j}")
        print("\n")

        final_segmentation_2, energy_2 = icm.icm_ising(
            negll, beta, initial_segmentation
        )
        initial_segmentation = final_segmentation_2

    difference_map = np.abs(final_segmentation_1 - final_segmentation_2)
    npt.assert_(not np.isclose(np.abs(np.sum(difference_map)), 0))


def test_classify():
    imgseg = TissueClassifierHMRF()

    nclasses = 4
    beta = 0.1
    tolerance = 0.0001
    max_iter = 10

    image = create_image()

    npt.assert_(image.max() == 1.0)
    npt.assert_(image.min() == 0.0)

    # First we test without setting iterations and tolerance
    seg_init, seg_final, PVE = imgseg.classify(image, nclasses, beta)

    npt.assert_(seg_final.max() == nclasses)
    npt.assert_(seg_final.min() == 0.0)

    # Second we test it with just changing the tolerance
    seg_init, seg_final, PVE = imgseg.classify(
        image, nclasses, beta, tolerance=tolerance
    )

    npt.assert_(seg_final.max() == nclasses)
    npt.assert_(seg_final.min() == 0.0)

    # Third we test it with just the iterations
    seg_init, seg_final, PVE = imgseg.classify(image, nclasses, beta, max_iter=max_iter)

    npt.assert_(seg_final.max() == nclasses)
    npt.assert_(seg_final.min() == 0.0)

    # Next we test saving the history of accumulated energies from ICM
    imgseg = TissueClassifierHMRF(save_history=True)

    seg_init, seg_final, PVE = imgseg.classify(
        200 * image, nclasses, beta, tolerance=tolerance
    )

    npt.assert_(seg_final.max() == nclasses)
    npt.assert_(seg_final.min() == 0.0)

    npt.assert_(imgseg.energies_sum[0] > imgseg.energies_sum[-1])



from os.path import join as pjoin
from tempfile import TemporaryDirectory
import warnings

import numpy as np
from numpy.testing import assert_allclose, assert_equal

from dipy.core.gradients import generate_bvecs
from dipy.data import get_fnames
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti, load_nifti_data, save_nifti
from dipy.io.peaks import load_pam
from dipy.reconst.shm import descoteaux07_legacy_msg
from dipy.workflows.reconst import ReconstDkiFlow


def test_reconst_dki():
    with TemporaryDirectory() as out_dir:
        data_path, bval_path, bvec_path = get_fnames(name="small_101D")
        volume, affine = load_nifti(data_path)
        mask = np.ones_like(volume[:, :, :, 0])
        mask_path = pjoin(out_dir, "tmp_mask.nii.gz")
        save_nifti(mask_path, mask.astype(np.uint8), affine)

        dki_flow = ReconstDkiFlow()

        args = [data_path, bval_path, bvec_path, mask_path]

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=descoteaux07_legacy_msg,
                category=PendingDeprecationWarning,
            )
            dki_flow.run(*args, out_dir=out_dir, extract_pam_values=True)

        fa_path = dki_flow.last_generated_outputs["out_fa"]
        fa_data = load_nifti_data(fa_path)
        assert_equal(fa_data.shape, volume.shape[:-1])

        tensor_path = dki_flow.last_generated_outputs["out_dt_tensor"]
        tensor_data = load_nifti_data(tensor_path)
        assert_equal(tensor_data.shape[-1], 6)
        assert_equal(tensor_data.shape[:-1], volume.shape[:-1])

        ga_path = dki_flow.last_generated_outputs["out_ga"]
        ga_data = load_nifti_data(ga_path)
        assert_equal(ga_data.shape, volume.shape[:-1])

        rgb_path = dki_flow.last_generated_outputs["out_rgb"]
        rgb_data = load_nifti_data(rgb_path)
        assert_equal(rgb_data.shape[-1], 3)
        assert_equal(rgb_data.shape[:-1], volume.shape[:-1])

        md_path = dki_flow.last_generated_outputs["out_md"]
        md_data = load_nifti_data(md_path)
        assert_equal(md_data.shape, volume.shape[:-1])

        ad_path = dki_flow.last_generated_outputs["out_ad"]
        ad_data = load_nifti_data(ad_path)
        assert_equal(ad_data.shape, volume.shape[:-1])

        rd_path = dki_flow.last_generated_outputs["out_rd"]
        rd_data = load_nifti_data(rd_path)
        assert_equal(rd_data.shape, volume.shape[:-1])

        mk_path = dki_flow.last_generated_outputs["out_mk"]
        mk_data = load_nifti_data(mk_path)
        assert_equal(mk_data.shape, volume.shape[:-1])

        ak_path = dki_flow.last_generated_outputs["out_ak"]
        ak_data = load_nifti_data(ak_path)
        assert_equal(ak_data.shape, volume.shape[:-1])

        rk_path = dki_flow.last_generated_outputs["out_rk"]
        rk_data = load_nifti_data(rk_path)
        assert_equal(rk_data.shape, volume.shape[:-1])

        kt_path = dki_flow.last_generated_outputs["out_dk_tensor"]
        kt_data = load_nifti_data(kt_path)
        assert_equal(kt_data.shape[-1], 15)
        assert_equal(kt_data.shape[:-1], volume.shape[:-1])

        mode_path = dki_flow.last_generated_outputs["out_mode"]
        mode_data = load_nifti_data(mode_path)
        assert_equal(mode_data.shape, volume.shape[:-1])

        evecs_path = dki_flow.last_generated_outputs["out_evec"]
        evecs_data = load_nifti_data(evecs_path)
        assert_equal(evecs_data.shape[-2:], (3, 3))
        assert_equal(evecs_data.shape[:-2], volume.shape[:-1])

        evals_path = dki_flow.last_generated_outputs["out_eval"]
        evals_data = load_nifti_data(evals_path)
        assert_equal(evals_data.shape[-1], 3)
        assert_equal(evals_data.shape[:-1], volume.shape[:-1])

        peaks_dir_path = dki_flow.last_generated_outputs["out_peaks_dir"]
        peaks_dir_data = load_nifti_data(peaks_dir_path)
        assert_equal(peaks_dir_data.shape[-1], 9)
        assert_equal(peaks_dir_data.shape[:-1], volume.shape[:-1])

        peaks_idx_path = dki_flow.last_generated_outputs["out_peaks_indices"]
        peaks_idx_data = load_nifti_data(peaks_idx_path)
        assert_equal(peaks_idx_data.shape[-1], 3)
        assert_equal(peaks_idx_data.shape[:-1], volume.shape[:-1])

        peaks_vals_path = dki_flow.last_generated_outputs["out_peaks_values"]
        peaks_vals_data = load_nifti_data(peaks_vals_path)
        assert_equal(peaks_vals_data.shape[-1], 3)
        assert_equal(peaks_vals_data.shape[:-1], volume.shape[:-1])

        pam = load_pam(dki_flow.last_generated_outputs["out_pam"])
        assert_allclose(pam.peak_dirs.reshape(peaks_dir_data.shape), peaks_dir_data)
        assert_allclose(pam.peak_values, peaks_vals_data)
        assert_allclose(pam.peak_indices, peaks_idx_data)

        bvals, bvecs = read_bvals_bvecs(bval_path, bvec_path)
        bvals[0] = 5.0
        bvecs = generate_bvecs(len(bvals))

        tmp_bval_path = pjoin(out_dir, "tmp.bval")
        tmp_bvec_path = pjoin(out_dir, "tmp.bvec")
        np.savetxt(tmp_bval_path, bvals)
        np.savetxt(tmp_bvec_path, bvecs.T)
        dki_flow._force_overwrite = True
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=descoteaux07_legacy_msg,
                category=PendingDeprecationWarning,
            )
            dki_flow.run(
                data_path,
                tmp_bval_path,
                tmp_bvec_path,
                mask_path,
                out_dir=out_dir,
                b0_threshold=0,
            )



"""Class to raise error for missing modules or other misfortunes"""


class TripWireError(AttributeError):
    """Exception if trying to use TripWire object"""


def is_tripwire(obj):
    """Returns True if `obj` appears to be a TripWire object

    Examples
    --------
    >>> is_tripwire(object())
    False
    >>> is_tripwire(TripWire('some message'))
    True
    """
    try:
        _ = obj.any_attribute
    except TripWireError:
        return True
    except Exception:
        pass
    return False


class TripWire:
    """Class raising error if used

    Standard use is to proxy modules that we could not import

    Examples
    --------
    >>> try:
    ...     import silly_module_name
    ... except ImportError:
    ...    silly_module_name = TripWire('We do not have silly_module_name')
    >>> silly_module_name.do_silly_thing('with silly string') #doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
        ...
    TripWireError: We do not have silly_module_name
    """  # noqa: E501

    def __init__(self, msg):
        self._msg = msg

    def __getattr__(self, attr_name):
        """Raise informative error accessing attributes"""
        raise TripWireError(self._msg)

    def __call__(self, *args, **kwargs):
        """Raise informative error while calling"""
        raise TripWireError(self._msg)
